// Copyright 1997
// College of Computer Science
// Northeastern University Boston MA 02115

// This software may be used for educational purposes as long as this copyright
// notice is retained at the top of all files

// Should this software be modified, the words "Modified from Original" must be
// included as a comment below this notice

// All publication rights are retained.  This software or its documentation may
// not be published in any media either in whole or in part.

///////////////////////////////////////////////////////////////////////////////

// Basic2D.cp
// Basic 2-dimensional types

///////////////////////////////////////////////////////////////////////////////

#include "Basic2D.h"
#include "IOTools.h"					// tools for safe I/O
#include "IOHelp.h"						// miscellaneous IO help routines


double AngleInRadians (const Vector2D& V, const Vector2D& W) {
	// easy case
	if (IsZero(V) || IsZero(W))
		return 0;
	
	// compute cosine of angle between vectors
	// arrange to compute only one sqrt
	
	double vv = DotProduct(V, V);
	double ww = DotProduct(W, W);
	double vw = DotProduct(V, W);
	
	double cosine = vw / sqrt(vv * ww);
	
	// check for round off on parallel vectors
	if (cosine >  1)
		cosine =  1;
	
	if (cosine < -1)
		cosine = -1;
	
	// compute angle using arc cosine function
	return acos( cosine );
}


void Print(ostream& os, const Vector2D& V, int width) {
	// x coordinate
	os << "X: ";
	
	if (width > 0)
		os << setw(width);
	
	os << V.x << endl;
	
	// y coordinate
	os << "Y: ";
	
	if (width > 0)
		os << setw(width);
	
	os << V.y << endl;
	
	// output blank line
	os << endl;
};


Vector2D RequestVector2D(			// with default
	const string&	prompt,			// prompt to user
	const Vector2D&	answer			// default answer
) {
	// output prompt and go to next line
	if (prompt.length())
		cout << prompt << endl;
	
	// get input
	double X;
	double Y;
	
	X = RequestDouble("X:", answer.x);
	Y = RequestDouble("Y:", answer.y);
	
	// output blank line
	cout << endl;
	
	// return result
	return Vector2D(X, Y);
}


Vector2D RequestVector2D(			// with no default
	const string&	prompt			// prompt to user
) {
	// output prompt and go to next line
	if (prompt.length())
		cout << prompt << endl;
	
	// get input
	double X;
	double Y;
	
	X = RequestDouble("X:");
	Y = RequestDouble("Y:");
	
	// output blank line
	cout << endl;
	
	// return result
	return Vector2D(X, Y);
}


bool ReadingVector2D(				// true if data provided
	const string&	prompt,			// prompt to user
	Vector2D&		result			// user data if provided
) {
	// output prompt and go to next line
	if (prompt.length())
		cout << prompt << endl;
	
	// get input
	double X;
	double Y;
	
	if (ReadingDouble("X:", X) && ReadingDouble("Y:", Y)) {
		result.Set(X, Y);

		// output blank line
		cout << endl;
		
		return true;
	}
	
	// output blank line
	cout << endl;
	
	// if reach here then no input so return false
	result.Set(0, 0);
	return false;
}


istream& operator>> (istream&is, Vector2D& V) {
	SkipToNumber(is);	// skip past non-numeric data such as "X: "
	is >> V.x;
	
	SkipToNumber(is);	// skip past non-numeric data such as "Y: "
	is >> V.y;
	
	return is;
}


void Print(ostream& os, const Point2D& P, int width) {
	// x coordinate
	os << "X: ";
	
	if (width > 0)
		os << setw(width);
	
	os << P.x << endl;
	
	// y coordinate
	os << "Y: ";
	
	if (width > 0)
		os << setw(width);
	
	os << P.y << endl;
	
	// output blank line
	os << endl;
}


Point2D RequestPoint2D(				// with default
	const string&	prompt,			// prompt to user
	const Point2D&	answer			// default answer
) {
	// output prompt and go to next line
	if (prompt.length())
		cout << prompt << endl;
	
	// get input
	double X;
	double Y;
	
	X = RequestDouble("X:", answer.x);
	Y = RequestDouble("Y:", answer.y);
	
	// output blank line
	cout << endl;
	
	// return result
	return Point2D(X, Y);
}


Point2D RequestPoint2D(				// with no default
	const string&	prompt			// prompt to user
) {
	// output prompt and go to next line
	if (prompt.length())
		cout << prompt << endl;
	
	// get input
	double X;
	double Y;
	
	X = RequestDouble("X:");
	Y = RequestDouble("Y:");
	
	// output blank line
	cout << endl;
	
	// return result
	return Point2D(X, Y);
}


bool ReadingPoint2D(				// true if data provided
	const string&	prompt,			// prompt to user
	Point2D&		result			// user data if provided
) {
	// output prompt and go to next line
	if (prompt.length())
		cout << prompt << endl;
	
	// get input
	double X;
	double Y;
	
	if (ReadingDouble("X:", X) && ReadingDouble("Y:", Y)) {
		result.Set(X, Y);

		// output blank line
		cout << endl;
		
		return true;
	}
	
	// output blank line
	cout << endl;
	
	// if reach here then no input so return false
	result.Set(0, 0);
	return false;
}


istream& operator>> (istream&is, Point2D& P) {
	SkipToNumber(is);	// skip past non-numeric data such as "X: "
	is >> P.x;
	
	SkipToNumber(is);	// skip past non-numeric data such as "Y: "
	is >> P.y;
	
	return is;
}


void Print(ostream& os, const Rect2D& R, int width) {
	// x1 coordinate
	os << "X1: ";
	
	if (width > 0)
		os << setw(width);
	
	os << R.x1 << endl;
	
	// y1 coordinate
	os << "Y1: ";
	
	if (width > 0)
		os << setw(width);
	
	os << R.y1 << endl;
	
	// x2 coordinate
	os << "X2: ";
	
	if (width > 0)
		os << setw(width);
	
	os << R.x2 << endl;
	
	// y2 coordinate
	os << "Y2: ";
	
	if (width > 0)
		os << setw(width);
	
	os << R.y2 << endl;
	
	// output blank line
	os << endl;
}

Rect2D RequestRect2D(				// with default
	const string&	prompt,			// prompt to user
	const Rect2D&	answer			// default answer
) {
	// output prompt and go to next line
	if (prompt.length())
		cout << prompt << endl;
	
	// get input
	double X1;
	double Y1;
	double X2;
	double Y2;
	
	X1 = RequestDouble("X1:", answer.x1);
	Y1 = RequestDouble("Y1:", answer.y1);
	X2 = RequestDouble("X2:", answer.x2);
	Y2 = RequestDouble("Y2:", answer.y2);
	
	// output blank line
	cout << endl;
	
	// return result
	return Rect2D(X1, Y1, X2, Y2);
}


Rect2D RequestRect2D(				// with no default
	const string&	prompt			// prompt to user
) {
	// output prompt and go to next line
	if (prompt.length())
		cout << prompt << endl;
	
	// get input
	double X1;
	double Y1;
	double X2;
	double Y2;
	
	X1 = RequestDouble("X1:");
	Y1 = RequestDouble("Y1:");
	X2 = RequestDouble("X2:");
	Y2 = RequestDouble("Y2:");
	
	// output blank line
	cout << endl;
	
	// return result
	return Rect2D(X1, Y1, X2, Y2);
}


bool ReadingRect2D(					// true if data provided
	const string&	prompt,			// prompt to user
	Rect2D&			result			// user data if provided
) {
	// output prompt and go to next line
	if (prompt.length())
		cout << prompt << endl;
	
	// get input
	double X1;
	double Y1;
	double X2;
	double Y2;
	
	if (
		ReadingDouble("X1:", X1) && ReadingDouble("Y1:", Y1)
		  &&
		ReadingDouble("X2:", X2) && ReadingDouble("Y2:", Y2)
	) {
		result.Set(X1, Y1, X2, Y2);

		// output blank line
		cout << endl;
		
		return true;
	}
	
	// output blank line
	cout << endl;
	
	// if reach here then no input so return false
	result.Set(0, 0, 0, 0);
	return false;
}


istream& operator>> (istream&is, Rect2D& R) {
	SkipToNumber(is);	// skip past non-numeric data such as "X1: "
	is >> R.x1;
	
	SkipToNumber(is);	// skip past non-numeric data such as "Y1: "
	is >> R.y1;

	SkipToNumber(is);	// skip past non-numeric data such as "X2: "
	is >> R.x2;
	
	SkipToNumber(is);	// skip past non-numeric data such as "Y2: "
	is >> R.y2;
	
	return is;
}


