// Copyright 1997
// College of Computer Science
// Northeastern University Boston MA 02115

// This software may be used for educational purposes as long as this copyright
// notice is retained at the top of all files

// Should this software be modified, the words "Modified from Original" must be
// included as a comment below this notice

// All publication rights are retained.  This software or its documentation may
// not be published in any media either in whole or in part.

///////////////////////////////////////////////////////////////////////////////

// Basic2D.h
// Basic 2-dimensional types

///////////////////////////////////////////////////////////////////////////////

#ifndef BASIC2D_H_
#define BASIC2D_H_

#include <iostream.h>					// stream definitions
#include "MathUtil.h"					// math & trig utilities


// forward declarations

class Vector2D;
class Point2D;
class Rect2D;


// declarations

class Vector2D {
	double x;		// x coordinate
	double y;		// y coordinate
	
public:
	friend class Point2D;
	friend class Rect2D;
	
	// default constructor
	
	Vector2D() : x(0), y(0) { };
	
	// explicit constructor
	
	Vector2D(double X, double Y) : x(X), y(Y) {  };
	
	// set operations
	
	Vector2D& Set(double X, double Y) { x = X; y = Y; return *this; };
	
	void SetX(double X) { x = X; };
	
	void SetY(double Y) { y = Y; };
	
	// get operation
	
	void Get(double& X, double& Y) const { X = x; Y = y; };
	
	// operator+	add vector V to vector W to get new vector V + W
	
	friend Vector2D operator+ (const Vector2D& V, const Vector2D& W) {
		return Vector2D( V.x + W.x, V.y + W.y );
	};
	
	// operator+=	add vector V to vector
	
	Vector2D& operator+= (const Vector2D& V) {
		x += V.x;
		y += V.y;
		
		return *this;
	};
	
	// operator-	subtract vector W from vector V to get new vector V - W
	
	friend Vector2D operator- (const Vector2D& V, const Vector2D& W) {
		return Vector2D( V.x - W.x, V.y - W.y );	
	};
	
	// operator-=	subtract vector V from vector
	
	Vector2D& operator-= (const Vector2D& V) {
		x -= V.x;
		y -= V.y;
		
		return *this;
	};
	
	// operator*	multiply vector V by scalar s to get new vector
	//				multiply may be done in either order V * s or s * V
	
	friend Vector2D operator* (const Vector2D& V, double s) {
		return Vector2D( V.x * s, V.y * s );
	};
	
	friend Vector2D operator* (double s, const Vector2D& V) {
		return Vector2D( s * V.x, s * V.y );
	};
	
	// operator*=	multiply vector by scalar s
	
	Vector2D& operator*= (double s) {
		x *= s;
		y *= s;
		
		return *this;
	};
	
	// operator/	divide vector V by scalar s to get new vector V / s
	
	friend Vector2D operator/ (const Vector2D& V, double s) {
		assert (s != 0);
		
		return Vector2D( V.x / s, V.y / s );
	};
	
	// operator/=	divide vector by scalar s
	
	Vector2D& operator/= (double s) {
		assert (s != 0);
		
		x /= s;
		y /= s;
		
		return *this;
	};
	
	// test if vector is zero
	
	friend bool IsZero (const Vector2D& V) {
		return (V.x == 0) && (V.y == 0);
	};
	
	// test ==
	
	friend bool operator== (const Vector2D& V, const Vector2D& W) {
		return IsZero(V - W);
	};
	
	// test !=
	
	friend bool operator!= (const Vector2D& V, const Vector2D& W) {
		return !(V == W);
	};
	
	// dot product
	
	friend double DotProduct (const Vector2D& V, const Vector2D& W) {
		return V.x * W.x + V.y * W.y;
	};
	
	// length of vector
	
	friend double Length (const Vector2D& V) {
		return sqrt(DotProduct(V, V));
	};
	
	// angles
	
	friend double AngleInRadians (const Vector2D& V, const Vector2D& W);
	
	friend double AngleInDegrees (const Vector2D& V, const Vector2D& W) {
		return radians_to_degrees * AngleInRadians(V, W);
	};
	
	// text output
	
	// text output via function call with optional field width
	friend void Print(ostream& os, const Vector2D& V, int width = 0);
	
	// text output via operator<<
	friend ostream& operator<< (ostream&os, const Vector2D& V) {
		Print(os, V);
		return os;
	};
	
	// text input
	
	friend Vector2D RequestVector2D(	// with default
		const string&	prompt,			// prompt to user
		const Vector2D&	answer			// default answer
	);
	
	friend Vector2D RequestVector2D(	// with no default
		const string&	prompt			// prompt to user
	);
	
	friend bool ReadingVector2D(		// true if data provided
		const string&	prompt,			// prompt to user
		Vector2D&		result			// user data if provided
	);

	// text input via operator>>
	// will read back the output of Print
	// will not do the error checking of IOTools
	friend istream& operator>> (istream&is, Vector2D& V);
};


class Point2D {
	double x;		// x coordinate
	double y;		// y coordinate
	
public:
	friend class Rect2D;
	
	// default constructor
	
	Point2D() : x(0), y(0) { };
	
	// explicit constructor
	
	Point2D(double X, double Y) : x(X), y(Y) {  };
	
	// set operations
	
	Point2D& Set(double X, double Y) { x = X; y = Y; return *this; };
	
	void SetX(double X) { x = X; };
	
	void SetY(double Y) { y = Y; };
	
	// get operation
	
	void Get(double& X, double& Y) const { X = x; Y = y; };
	
	// operator+	add vector V to point P to get point P + V
	
	friend Point2D operator+ (const Point2D& P, const Vector2D& V) {
		return Point2D( P.x + V.x, P.y + V.y);
	};
	
	// operator+=	add vector V to point
	
	Point2D& operator+= (const Vector2D& V) {
		x += V.x;
		y += V.y;
		
		return *this;
	};
	
	// operator-	subtract vector V from point P to get point P - V
	
	friend Point2D operator- (const Point2D& P, const Vector2D& V) {
		return Point2D( P.x - V.x, P.y - V.y);
	};
	
	// operator-=	subtract vector V from point
	
	Point2D& operator-= (const Vector2D& V) {
		x -= V.x;
		y -= V.y;
		
		return *this;
	};
	
	// operator-	subtract point Q from point P to get vector P - Q
	
	friend Vector2D operator- (const Point2D& P, const Point2D& Q) {
		return Vector2D( P.x - Q.x, P.y - Q.y );
	};
	
	// test ==
	friend bool operator== (const Point2D& P, const Point2D& Q) {
		return IsZero(P - Q);
	};
	
	// test !=
	friend bool operator!= (const Point2D& P, const Point2D& Q) {
		return !(P == Q);
	};
	
	// find distance from P to Q
	
	friend double Distance (const Point2D& P, const Point2D& Q) {
		return Length(P - Q);
	};

	// text output
	
	// text output via function call with optional field width
	friend void Print(ostream& os, const Point2D& P, int width = 0);
	
	// text output via operator<<
	friend ostream& operator<< (ostream&os, const Point2D& P) {
		Print(os, P);
		return os;
	};
	
	// text input
	
	friend Point2D RequestPoint2D(		// with default
		const string&	prompt,			// prompt to user
		const Point2D&	answer			// default answer
	);
	
	friend Point2D RequestPoint2D(		// with no default
		const string&	prompt			// prompt to user
	);
	
	friend bool ReadingPoint2D(			// true if data provided
		const string&	prompt,			// prompt to user
		Point2D&		result			// user data if provided
	);

	// text input via operator>>
	// will read back the output of Print
	// will not do the error checking of IOTools
	friend istream& operator>> (istream&is, Point2D& P);
	
	// conversions
	
	friend Vector2D Point2DtoVector2D(const Point2D& P) {
		return Vector2D(P.x, P.y);
	};
	
	friend Point2D Vector2DtoPoint2D(const Vector2D& V) {
		return Point2D(V.x, V.y);
	};
};


class Rect2D {
	double x1;		// x coordinate of corner 1
	double y1;		// y coordinate of corner 1
	double x2;		// x coordinate of corner 2
	double y2;		// y coordinate of corner 2
	
public:

	Rect2D() : x1(0), y1(0), x2(0), y2(0) { };
	
	Rect2D(double X1, double Y1, double X2, double Y2) :
		x1(X1), y1(Y1), x2(X2), y2(Y2) { };
	
	Rect2D(const Point2D& P1, const Point2D& P2) :
		x1(P1.x), y1(P1.y), x2(P2.x), y2(P2.y) { };
	
	// set operations
	
	Rect2D& Set(double X1, double Y1, double X2, double Y2) {
		x1 = X1;
		y1 = Y1;
		x2 = X2;
		y2 = Y2;
		return *this;
	};
	
	Rect2D& Set(const Point2D& P1, const Point2D& P2) {
		P1.Get(x1, y1);
		P2.Get(x2, y2);
		return *this;
	};
	
	// get operations
	
	void Get(double& X1, double& Y1, double& X2, double& Y2) const {
		X1 = x1;
		Y1 = y1;
		X2 = x2;
		Y2 = y2;
	};
	
	void Get(Point2D& P1, Point2D& P2) const {
		P1.Set(x1, y1);
		P2.Set(x2, y2);
	};
	
	// shift corners of rectangle using Vector2D V
	
	void Shift(const Vector2D& V) {
		x1 += V.x;		// add
		y1 += V.y;		// add
		
		x2 += V.x;		// add
		y2 += V.y;		// add
	};
	
	// inset corners of rectangle using Vector2D V
	
	void Inset(const Vector2D& V) {
		x1 += V.x;		// add
		y1 += V.y;		// add
		
		x2 -= V.x;		// subtract
		y2 -= V.y;		// subtract
	};
	
	// expand rectangle to include Point2D P
	
	void Expand(const Point2D& P) {
		ExpandInterval(x1, x2, P.x);
		ExpandInterval(y1, y2, P.y);
	};
	
	// area
	
	friend double Area (const Rect2D& R) {
		return fabs( (R.x2 - R.x1) * (R.y2 - R.y1) );
	};

	
	// perimeter
	
	friend double Perimeter(const Rect2D& R) {
		return 2 * ( fabs(R.x2 - R.x1) + fabs(R.y2 - R.y1) );
	};

	// text output
	
	// text output via function call with optional field width
	friend void Print(ostream& os, const Rect2D& R, int width = 0);
	
	// text output via operator<<
	friend ostream& operator<< (ostream&os, const Rect2D& R) {
		Print(os, R);
		return os;
	};
	
	// text input
	
	friend Rect2D RequestRect2D(		// with default
		const string&	prompt,			// prompt to user
		const Rect2D&	answer			// default answer
	);
	
	friend Rect2D RequestRect2D(		// with no default
		const string&	prompt			// prompt to user
	);
	
	friend bool ReadingRect2D(			// true if data provided
		const string&	prompt,			// prompt to user
		Rect2D&			result			// user data if provided
	);

	// text input via operator>>
	// will read back the output of Print
	// will not do the error checking of IOTools
	friend istream& operator>> (istream&is, Rect2D& R);
};


#endif // BASIC2D_H_