// Copyright 1996
// College of Computer Science
// Northeastern University Boston MA 02115

// This software may be used for educational purposes as long as this copyright
// notice is retained at the top of all files

// Should this software be modified, the words "Modified from Original" must be
// included as a comment below this notice

// All publication rights are retained.  This software or its documentation may
// not be published in any media either in whole or in part.

///////////////////////////////////////////////////////////////////////////////

// ChBuffer.h

#ifndef CHBUFFER_H_
#define CHBUFFER_H_

#include "CHeaders.h"


// Character buffer template class

enum {
SmallBufferSize =  32,
LargeBufferSize = 256
};


template <long BufferSize> class CharBuffer {
	long	size;
	char	buffer[BufferSize];
	
	// SetNull guarantees that buffer is null terminated
	void SetNull(void) {
		buffer[size - 1] = '\0';
	};
	
public:
	// default initialization to null string
	
	CharBuffer() : size(BufferSize) {
		memset(buffer, '\0', size);
	};
	
	// initialization with given string data
	// will truncate string if necessary
	
	CharBuffer(const char* string) : size(BufferSize) {
		int length = strlen(string);
		
		if (length >= size)
			length = size - 1;
		
		if (length > 0)
			memmove(buffer, string, length);
		
		memset(buffer + length, '\0', size - length);	
	};
	
	// operator = for a string
	// will truncate string if necessary
	
	CharBuffer& operator= (const char* string) {
		int length = strlen(string);
		
		if (length >= size)
			length = size - 1;
		
		if (length > 0)
			memmove(buffer, string, length);
		
		memset(buffer + length, '\0', size - length);
		
		return *this;	
	};
	
	// append operator << for a string
	// will truncate string if necessary
	
	CharBuffer& operator<< (const char* string) {
		SetNull();
		
		int offset = strlen(buffer);
		int room   = size - offset;
		int length = strlen(string);
		
		char* location = buffer + offset;
		
		if (length >= room)
			length = room - 1;
		
		if (length > 0)
			memmove(location, string, length);
		
		memset(location + length, '\0', room - length);
		
		return *this;	
	};
	
	// insert substring from start to next into CharBuffer
	
	CharBuffer& Insert(const char* start, const char* next) {
		int length = next - start;
		
		if (length >= size)
			length = size - 1;
		
		if (length > 0)
			memmove(buffer, start, length);
		
		memset(buffer + length, '\0', size - length);
		
		return *this;	
	};
	
	// append substring from start to next into CharBuffer
	
	CharBuffer& Append(const char* start, const char* next) {
		SetNull();
		
		int offset = strlen(buffer);
		int room   = size - offset;
		int length = next - start;
		
		char* location = buffer + offset;
		
		if (length >= room)
			length = room - 1;

		if (length > 0)
			memmove(location, start, length);
		
		memset(location + length, '\0', room - length);
		
		return *this;	
	};
	
	// copy constructor for a CharBuffer
	// will truncate B's data if necessary
	
	CharBuffer(const CharBuffer& B) : size(BufferSize) {
		int length = B.size;
		
		if (length >= size)
			length = size - 1;
		
		memmove(buffer, B.buffer, length);
		
		memset(buffer + length, '\0', size - length);	
	};
	
	// operator = for a CharBuffer
	// will truncate B's data if necessary
	
	CharBuffer& operator= (const CharBuffer& B) {
		int length = B.size;
		
		if (length >= size)
			length = size - 1;
		
		memmove(buffer, B.buffer, length);
		
		memset(buffer + length, '\0', size - length);
		
		return *this;
	};
	
	// append operator << for a CharBuffer
	// will truncate B's data if necessary
	
	CharBuffer& operator<< (const CharBuffer& B) {
		SetNull();
		
		int offset = strlen(buffer);
		int room   = size - offset;
		int length = B.size;
		
		char* location = buffer + offset;
		
		if (length >= room)
			length = room - 1;
		
		if (length > 0)
			memmove(location, B.buffer, length);
		
		memset(location + length, '\0', room - length);
		
		return *this;	
	};
	
	// automatic conversion back to standard char pointer
	// we will permit buffer to be modified via the pointer
	//
	// this operator is critical for output and for sending
	// data to character buffers of different sizes
	//
	// we set buffer[size - 1] to null before return
	
	operator char* () {
		SetNull();
		
		return buffer;
	};
	
	// operator[]
	// if index < 0			return buffer[0]
	// if index >= size		return buffer[size - 1]
	//
	// in effect we choose to return a valid char reference
	// rather than force a crash when index is out of range
	//
	// we set buffer[size - 1] to null before return
	
	char& operator[] (long index) {
		SetNull();
		
		if (index < 0)
			index = 0;
		else if (index >= size)
			index = size - 1;
		
		return buffer[index];
	}
	
	// return size of buffer
	
	long Size(void) const {
		return size;
	};
};


typedef CharBuffer<SmallBufferSize> SmallBuffer;

typedef CharBuffer<LargeBufferSize> LargeBuffer;


#endif // CHBUFFER_H_

