// Copyright 1997
// College of Computer Science
// Northeastern University Boston MA 02115

// This software may be used for educational purposes as long as this copyright
// notice is retained at the top of all files

// Should this software be modified, the words "Modified from Original" must be
// included as a comment below this notice

// All publication rights are retained.  This software or its documentation may
// not be published in any media either in whole or in part.

///////////////////////////////////////////////////////////////////////////////

// Delay.cp

///////////////////////////////////////////////////////////////////////////////

#include "Delay.h"
#include "IOTools.h"

long DelayTicks		= 0;		// default is no delay

bool ManualMode		= false;	// default is no pause

bool OnButtonFreeze	= false;	// default is no freeze on mouse press

// flags for actual key states

bool CommandFlag	= false;
bool OptionFlag		= false;
bool ControlFlag	= false;
bool ShiftFlag		= false;
bool CapsLockFlag	= false;

// flags to signal action

bool BreakFlag		= false;	// convention: true if command down
bool SpeedFlag		= false;	// convention: true if caps lock down
bool AbortFlag		= false;	// convention: true if command-option-control down


// delay routines

void Delay(long ticks) {
	// exit immediately if on entry:
	//		caps lock down and mouse up
	//		command-option-control down
	
	ResetFlags();						// reset flags
	
	if (SpeedFlag) {					// here caps lock down
	
		// check once for mouse down before immediate return
		
		if (Button()) {					// if notice mouse down
			if (OnButtonFreeze) {		// if freeze on mouse down
				AwaitRelease();			// pause until mouse up
				ResetFlags();			// reset flags again
			}
		}
		
		return;							// then exit
	}
	
	// compute time to finish
	clock_t	finish = clock() + (CLOCKS_PER_SEC/60) * ticks;
	
	// waste time until finish reached
	do {
		// check continuously for mouse down while wasting time
		
		if (Button()) {					// if notice mouse down
			if (OnButtonFreeze)			// if freeze on mouse down
				AwaitRelease();			// pause until mouse up

			ResetFlags();				// reset flags again
			return;						// then exit
		}
	} while (clock() < finish);
	
	ResetFlags();						// reset flags again
}


void SetDelay(long ticks) {
	if (ticks < 0)
		ticks = 0;
	
	DelayTicks = ticks;
}


long GetDelay() {
	return DelayTicks;
}


void RequestDelay(const string& prompt, long answer) {
	long ticks = RequestLong(prompt, answer);
	cout << endl;
	
	SetDelay(ticks);
}


// pause routines

void Pause() {
	if (ManualMode)
		AwaitClick();
	else
		Delay();
}


void SetPause(bool mode) {
	ManualMode = mode;
}


bool GetPause() {
	return ManualMode;
}


void RequestPause(const string& prompt, bool answer) {
	ManualMode = Confirm(prompt, answer);
	
	if (!ManualMode)
		RequestDelay();
}


// mouse utilities

void AwaitPress() {
	ResetFlags();					// reset flags
	
	if (SpeedFlag)
		return;
	
	while (!Button()) { };			// wait for mouse down
	
	ResetFlags();					// reset flags again
}


void AwaitRelease() {
	ResetFlags();					// reset flags
	
	while (Button()) { };			// wait for mouse up	
	
	ResetFlags();					// reset flags again
}


void AwaitClick() {
	AwaitPress();
	AwaitRelease();
}


// key press utilities

short ModifierStatus() {
	// See Inside Macintosh I (251-252, 259-260) and V (191-194)
	//		or
	// See Inside Macintosh: Macintosh Toolbox Essentials: Chapter 2
	//		2-38 to 2-42 and 2-110 and C Summary 2-125 to 2-129
	//		Warning: On 2-110 ignore phrase "minus 1" - it is wrong!
	
	KeyMap	keys;		// Read keyboard state
	GetKeys(keys);
	
	// the data on the key press status of special keys is found in:
	//		sequence bits 37 to 3B (hex) of keys array, that is,
	//		sequence bits 17 to 1B (hex) of keys[1]
	//
	// the arithmetic bit ordering in a long is as follows:
	//		1F ... 18 | 17 ... 10 | 0F ... 08 | 07 ... 00
	//
	// the leftmost byte 1F ... 18 is the lowest order byte arithmetically!
	// the next byte 17 ... 10 is the 2nd lowest order byte arithmetically!
	//
	// thus viewed arithmetically as a long rather than as a bit sequence:
	//		1B .. 18 are the four lowest bits of the lowest byte!
	//		17 is the highest bit of the 2nd lowest byte!
	//
	// thus bit sequence ordering is different from arithmetic bit ordering
	//
	// in fact:
	//		sequence bit 18 ==>		arithmetic bit 0
	//		sequence bit 19 ==>		arithmetic bit 1
	//		sequence bit 1A ==>		arithmetic bit 2
	//		sequence bit 1B ==>		arithmetic bit 3
	//		sequence bit 17 ==>		arithmetic bit F
	// 
	// the key masks are defined as follows:
	//		cmdKey		 256		arithmetic bit 8
	//		shiftKey	 512		arithmetic bit 9
	//		alphaLock	1024		arithemtic bit A
	//		optionKey	2048		arithemtic bit B
	//		controlKey	4096		arithemtic bit C
	//
	// thus we are forced to do bit operations to move around the status bits
	// to match the conventions of the key masks
	//
	// it is important to use long rather than short in calculations since if
	// x2 is short then bit is sign bit and 0x1000 >>7 will produce 0xFF00
	
	long x1 = keys[1] & 0x000F;			// Get low 4 bits of low word of keys[1]
	long x2 = keys[1] & 0x8000;			// Get high bit of low word of keys[1]
	long xx = (x1 << 9) | (x2 >> 7);	// Shift and combine
	
	return short(xx);
}


void ResetFlags() {
	short status = ModifierStatus();
	
	// flags for actual key states
	
	CommandFlag		= ((status & cmdKey)		== cmdKey);
	OptionFlag		= ((status & optionKey)		== optionKey);
	ControlFlag		= ((status & controlKey)	== controlKey);
	ShiftFlag		= ((status & shiftKey)		== shiftKey);
	CapsLockFlag	= ((status & alphaLock)		== alphaLock);
	
	// flags to signal action
	
	BreakFlag = CommandFlag;
	
	SpeedFlag = CapsLockFlag;
	
	AbortFlag = CommandFlag && OptionFlag && ControlFlag;
	
	if (AbortFlag)						// if user wants abort
		throw UserAbort();				// throw exception to exit rapidly
}
