// Copyright 1997
// College of Computer Science
// Northeastern University Boston MA 02115

// This software may be used for educational purposes as long as this copyright
// notice is retained at the top of all files

// Should this software be modified, the words "Modified from Original" must be
// included as a comment below this notice

// All publication rights are retained.  This software or its documentation may
// not be published in any media either in whole or in part.

///////////////////////////////////////////////////////////////////////////////

// Graphics.cp

#include "Graphics.h"
#include "IOTools.h"

// Tools for RGB string initialization and input

RGB RequestRGB(			// with default
	const string&	prompt,			// prompt to user
	const RGB		answer			// default answer
) {
	RGB rgb;
	
	string stringanswer(answer.RGBToString());
	
	string response;
	
	while (true) {		
		RequestString(prompt, stringanswer, response);
		
		if (InputError() || response.length() == 0)
			return answer;
		
		int start = 0;
		int next  =	0;
		int error = NoError;
		
		rgb = StringToRGB(response, start, next, error);
		
		if (!error)
			return rgb;
		
		ErrorPrint(response, next, error);
	}
}


RGB RequestRGB(			// without default
	const string&	prompt			// prompt to user
) {
	RGB rgb;
	
	string response;
	
	while (true) {		
		RequestString(prompt, response);
		
		if (InputError())
			return rgb.Set(0,0,0);
		
		int start = 0;
		int next  =	0;
		int error = NoError;
		
		if (response.length()) {
			rgb = StringToRGB(response, start, next, error);
			
			if (!error)
				return rgb;
		}
		else
			error = NoDataError;
		
		ErrorPrint(response, next, error);
	}
}


bool ReadingRGB(
	const string&	prompt,			// prompt to user
	RGB&			answer			// user answer if provided
) {
	string response;
	
	while (true) {		
		RequestString(prompt, response);
		
		if (InputError() || response.length() == 0) {
			answer.Set(0,0,0);
			return false;
		}
		
		int start = 0;
		int next  =	0;
		int error = NoError;
		
		answer = StringToRGB(response, start, next, error);
		
		if (!error)
			return true;
		
		ErrorPrint(response, next, error);
	}
}


RGB StringToRGB(
	const string&	source,			// string to read
	int				start,			// start position to read
	int&			next,			// next  position to read
	int&			error			// string IO error code
) {
	// define return value and initialize to default of black
	RGB rgb;
	
	// initialize error
	error = NoError;
	
	// find size
	int size = source.length();
	
	// check start
	if (start < 0)
		start = 0;
	
	if (start >= size) {
		next  = size;
		error = NoDataError;
		return rgb;
	}
	
	// process string via const char* pointer
	
	const char* temp = source.c_str() + start;
	const char* stop = temp;
	
	// use C string version of StringToRGB
	
	rgb = StringToRGB(temp, stop, error);
	
	// clean up
	next = start + (stop - temp);
	
	return rgb;
}


RGB StringToRGB(
	const char*		start,			// start position
	const char*&	next,			// next position to read
	int&			error			// error code
) {
	SkipSpace(start);
	next	= start;
	error	= NoError;
		
	byte r = 0;						// color components
	byte g = 0;
	byte b = 0;
	
	int count = 0;					// component number
	
	while (count < 3) {
		// Count which field
		count++;
		
		// Parse field value
		if (isdigit(*next)) {
			errno = 0;				// reset ANSI error state
			
			char* t = (char*) next;	// copy next to t
			
			unsigned long data = strtoul(next, &t, 0);	// convert
			
			next = t;				// copy t to next
			
			if (errno) {
				error = ColorError;		// Error in color byte
				break;
			}
			
			if (data > 255) {
				error = ColorError;		// Error in color byte
				break;
			}
			
			// Assign color to field
			
			switch (count) {
				case 1:
					r = data;
					break;
				
				case 2:
					g =	data;
					break;
				
				case 3:
					b =	data;
					break;
				
				default:
					break;
			}
			
			// Remove comma separator if present
			
			if (count < 3) {
				SkipSpace(next);
			
				if (*next == Comma) {
					next++;
					SkipSpace(next);
				}
			}
		}
		else if (*next == Null) {
			error = EndDataError;			// Not enough data
			break;
		}
		else {
			error = BadDataError;			// Garbage
			break;
		}
	}
	
	return RGB(r, g, b);
}


string RGB::RGBToString(bool wide) const {
	// define stream
	stringstream stream;
	
	// format to the stream
	if (wide)
		stream << right
		       << setw(3) << color[0] << ", "
		       << setw(3) << color[2] << ", " 
		       << setw(3) << color[4] << ends;
	else
		stream << color[0] << ", "
		       << color[2] << ", " 
		       << color[4] << ends;
	
	// copy stream to string
	return stream.str();
}