// Copyright 1997
// College of Computer Science
// Northeastern University Boston MA 02115

// This software may be used for educational purposes as long as this copyright
// notice is retained at the top of all files

// Should this software be modified, the words "Modified from Original" must be
// included as a comment below this notice

// All publication rights are retained.  This software or its documentation may
// not be published in any media either in whole or in part.

///////////////////////////////////////////////////////////////////////////////

// IOTools.h

///////////////////////////////////////////////////////////////////////////////

// IOTools has the following purposes:
//
// 1. To provide safe (error checked) interactive input that permits a user to
//    re-enter data in which errors are detected.
//
// 2. To automate the prompt-response cycle so that "io clutter" in the caller
//    program is minimized.
//
// 3. To enable the use of default responses if desired.
//
// 4. To integrate io into control flow smoothly.  For example to permit loops
//    of the form:
//
//		while (ReadingString(myprompt, mystring)) { ... }
//
//		while (ReadingDouble(myprompt, mydouble)) { ... }
//
// 5. To simplify conversions between strings and numbers.
//
// 6. Define a framework that will permit safe-io to be extended to user types.

///////////////////////////////////////////////////////////////////////////////

// Changes: Summer 1997

// The original version of IOTools used a simple template-based "string class"
// that was tightly integrated with classic pointer-based C strings.  This
// class is defined in ChBuffer.h and its header begins:
//
//		template <long BufferSize> class CharBuffer {
//			long	size;
//			char	buffer[BufferSize];
//
// Here, BufferSize is the fixed amount of space to allocate for the character
// data and size is a local copy of this initialization value.  Two particular
// instances of this CharBuffer class are used extensively in existing course
// materials:
//
//		typedef CharBuffer<32>  SmallBuffer
//		typedef CharBuffer<256> LargeBuffer

// To a substantial degree, the CharBuffer class has been made obsolete by the
// standard string class now available in C++.  It is our intention to update
// all educational materials to use standard strings but it was felt that this
// change will take time and that backward compatibility is essential at this
// stage.  Therefore we will define multiple versions for some function calls.

// Now that direct support for type bool is available, we no longer need to use
// the header file bool.h.

// There is one call that has been made entirely obsolete by calls in <iomanip>:
//
//		void ConsoleTextJustify(int how);
//
// This function is now omitted and therefore the header Enum.h need not be used
// in IOTools.

///////////////////////////////////////////////////////////////////////////////

#ifndef IOTOOLS_H_
#define IOTOOLS_H_

// #include <bool.h>					Omit!
// #include "Enum.h"					Omit!

#include <string>						// C++ string class definition
#include "MathUtil.h"					// For mathematical stuff
#include "ChBuffer.h"					// Retain for backward compatibility

///////////////////////////////////////////////////////////////////////////////

// Preliminary definitions

// typedef for functions to test conditions on characters
// to comply with C library functions that use int-to-int

typedef int (*IntFcnPtr) (int);


// Useful character constants
	
const char	BackSlash	= '\\';
const char	Blank		= ' ';
const char	Comma		= ',';
const char	Tab			= '\t';
const char	NL			= '\n';
const char	CR			= '\r';
const char	FF			= '\f';
const char	VT			= '\v';
const char	Null		= '\0';


// Case conversion declarations

enum CaseAction {
	// The following case actions can be requested as the global default
	// case action in SetCaseAction:

	None,					// No case conversion
	LowerCase,				// Convert to lower case
	UpperCase,				// Convert to upper case
	MixedCase				// Capitalize the first letter in every word &
							// convert the rest of each word to lower case
};

void SetCaseAction(CaseAction  action);		// Set global case action

void GetCaseAction(CaseAction& action);		// Look up global case action

///////////////////////////////////////////////////////////////////////////////

// IOTools can read one or more lines of user input into a character buffer
// or string subject to the following conventions:
//
//     1. If a line ends with  \  then the next line will be joined to it
//        and the backslash will be deleted from the input stream
//
//     2. Leading and trailing whitespace will be deleted from the buffer
//        once a line is read that does not end in a backslash
//
//     3. Default case conversion may be specified using SetCaseAction
//            None         No case conversion
//            LowerCase    Convert to lower case
//            UpperCase    Convert to upper case
//            MixedCase    Capitalize the first letter in every word &
//                         convert the rest of each word to lower case
//
//     4. If input begins with  \  then case conversion will be not be done
//        and the backslash will be deleted from the input stream
//
//     5. A prompt string may be specified
//
//     6. A default answer string may be specified

///////////////////////////////////////////////////////////////////////////////

// PromptAnswer handles the initial input phase in which a prompt and a default
// answer are provided to the user.  This tool is used throughout IOTools and
// is made public so that programmers can easily adapt IOTools to their own
// types.

void PromptAnswer(
	const string&	prompt,				// prompt to user
	const string&	answer				// default answer
);

///////////////////////////////////////////////////////////////////////////////

// RequestString is the basic tool on which all other reading tools are based

///////////////////////////////////////////////////////////////////////////////

// RequestString using the C++ string class has three forms:
//
//		void RequestString(
//			const string&	prompt,				// prompt to user
//			const string&	answer,				// default answer
//			string& 		response			// user response
//		);
//
//		void RequestString(
//			const string&	prompt,				// prompt to user
//			string& 		response			// user response
//		);
//
//		void RequestString(
//			string& 		response			// user response
//		);

// The string version of RequestString returns void to emphasize that the data
// is returned in the reference parameter response.  This avoids an extra copy
// of the data that would be required if the string data was returned by value.

// Description of RequestString:
//
// Parameter  prompt  supplies a user prompt string.  If this parameter is
// omitted, then there is no prompt printed
//
// Parameter  answer  is the default data returned if the user presses RETURN
// without giving input.  If this parameter is omitted, then the empty string
// is the default.
//
// The data read is copied into the parameter  response.
//
// Before the string data is returned:
//
//     Leading and trailing whitespace is deleted
//
//     The current global case action is performed unless overridden by
//     a backslash at the start of the input string

void RequestString(
	const string&	prompt,				// prompt to user
	const string&	answer,				// default answer
	string& 		response			// user response
);

inline void RequestString(
	const string&	prompt,				// prompt to user
	string& 		response			// user response
) {
	RequestString(prompt, "", response);
}

inline void RequestString(
	string& 		response			// user response
) {
	RequestString("", "", response);
}

///////////////////////////////////////////////////////////////////////////////

// RequestString using CharBuffer tools:
//
// 		char* RequestString(
//			const char*		prompt = "",		// prompt to user
//			const char*		answer = "",		// default answer
//			LargeBuffer& 	buffer = IOBuffer	// data buffer
//		);
//
// The default buffer is defined here:

extern LargeBuffer IOBuffer;		// Default input buffer for RequestString 

// Description of RequestString:
//
// Parameter  prompt  supplies an optional user prompt string
//
// Parameter  answer  is the default data returned if the user presses
// the return key without giving input
//
// The data read is copied into the data area specified by parameter  buffer
// IOBuffer is the default buffer
//
// The address of  buffer  is the return value of RequestSting so that other
// C-style routines can use the string data stored
//
// This permits (not so safe) calls of the form:
//
//     char* s;
//     s = RequestString(...);
//
// in which the user ignores the name or location of the default buffer.  As
// in C, the data is not moved but stays in IOBuffer.  We permit the pointer
// s to point directly at this data.  Of course, after a subsequent call to
// RequestString, the original data will be overwritten so s will appear to
// point to "new data".
//
// To read and copy data safely, make calls of the form:
//
//     enum { MySize = ... };			// Choose size of my buffer
//     CharBuffer<MySize>	MyBuffer;	// Define my buffer
//     MyBuffer = RequestString(...);	// Send data into my buffer
//
// The data from the buffer parameter is then transferred into MyBuffer using
// a safe overloaded version of operator= for the CharBuffer class.  The data
// will be truncated if it contains more than MySize - 1 valid characters.
//
// Before the string data is returned:
//
//     Leading and trailing whitespace is deleted
//
//     The current global case action is performed unless overridden by
//     a backslash at the start of the input string

char* RequestString(
	const char*		prompt = "",		// prompt to user
	const char*		answer = "",		// default answer
	LargeBuffer& 	buffer = IOBuffer	// data buffer
);

///////////////////////////////////////////////////////////////////////////////

// ReadingString takes a prompt and a string-like reference parameter.
// The string-like parameter can either be a C++ string object a LargeBuffer.
// ReadingString returns the number of valid characters input.

///////////////////////////////////////////////////////////////////////////////

int ReadingString(
	const string&	prompt,				// prompt to user
	string& 		response			// user response
);

int ReadingString(
	const char*		prompt,				// prompt to user
	LargeBuffer&	buffer				// buffer for user answer
);

///////////////////////////////////////////////////////////////////////////////

// Check for input error on cin

bool InputError();

///////////////////////////////////////////////////////////////////////////////

// Useful string manipulations

///////////////////////////////////////////////////////////////////////////////

// Versions for classic C strings

// Define tool to automatically change case

void ChangeCase(char* s, CaseAction action = None);

// Define squash function to delete leading and trailing whitespace in s

void SquashString(char* s);

// Skip white space in s by updating s to next non-white-space character

void SkipSpace(const char*& s);

// Skip characters in s while Test function is non-zero
// Stop if null character is reached

void SkipWhile(const char*& s, IntFcnPtr Test);

///////////////////////////////////////////////////////////////////////////////

// Versions for C++ strings

// Define tool to automatically change case

void ChangeCase(string& s, CaseAction action = None);

// Define squash function to delete leading and trailing whitespace in s

void SquashString(string& s);


// Since C++ strings are self-contained we can have both forward and reverse
// variants of the Skip functions.


// Skip white space in s
//
// Begin at position start and end at position stop
//
// Require:
//		0 <= start
// If this condition fails then set start = 0.
//
// Require:
//		start < s.length()
// If this condition fails then set stop = s.length() and return immediately.
//
// Otherwise return stop such that:
//		start <= stop <= s.length()
//		if stop < s.length() then:
//				s[stop] is first non-space equal to or after s[start]

void SkipSpace(const string& s, int start, int& stop);


// Skip white space in s in reverse direction
//
// Begin at position start and end at position stop
//
// Require:
//		start < s.length()
// If this condition fails then set start = s.length() - 1.
//
// Require:
//		0 <= start
// If this condition fails then set stop = -1 and return immediately.
//
// Otherwise return stop such that:
//		-1 <= stop <= start
//		if 0 <= stop then:
//				s[stop] is first non-space equal to or before s[start]

void ReverseSkipSpace(const string& s, int start, int& stop);


// Skip characters in s while Test function is non-zero
//
// Begin at position start and end at position stop
//
// Require:
//		0 <= start
// If this condition fails then set start = 0.
//
// Require:
//		start < s.length()
// If this condition fails then set stop = s.length() and return immediately.
//
// Otherwise return stop such that:
//		start <= stop <= s.length()
//		for all i such that start <= i < stop:   Test(s[i])    != 0
//		if stop < s.length() then:               Test(s[stop]) == 0

void SkipWhile(const string& s, IntFcnPtr Test, int start, int& stop);


// Skip characters in s in reverse direction while Test function is non-zero
//
// Begin at position start and end at position stop
//
// Require:
//		start < s.length()
// If this condition fails then set start = s.length() - 1.
//
// Require:
//		0 <= start
// If this condition fails then set stop = -1 and return immediately.
//
// Otherwise return stop such that:
//		-1 <= stop <= start
//		for all i such that stop < i <= start:   Test(s[i])    != 0
//		if 0 <= stop then:                       Test(s[stop]) == 0

void ReverseSkipWhile(const string& s, IntFcnPtr Test, int start, int& stop);

///////////////////////////////////////////////////////////////////////////////

// PressReturn tool

// Print a newline
// If message is not empty then print the message and a newline
// Print "Press Return To Continue "
// Pause until user presses return and eat any miscellaneous input
// Print a newline

void PressReturn(const string& message = "");

///////////////////////////////////////////////////////////////////////////////

// Confirm tool

// Generic tool to ask Yes-No questions
// Y or Yes corresponds to true
// N or No  corresponds to false

bool Confirm(
	const string&	prompt,			// prompt to user
	bool			answer			// default answer
);

///////////////////////////////////////////////////////////////////////////////

// Single character input

// Version 1: Supply default

char RequestChar(
	const string&	prompt,			// prompt to user
	char			answer			// default answer
);

// Version 2: Demand answer - no default

char RequestChar(
	const string&	prompt			// prompt to user
);

// Version 3:
//		Return data read in a parameter
//		Return true as function value if data entered
//		by user - otherwise return false

bool ReadingChar(
	const string&	prompt,			// prompt to user
	char&			answer			// user answer
);

///////////////////////////////////////////////////////////////////////////////

// Numeric data input
//
// For each numeric type  T  there are four functions
//
//		T RequestT(			// with default
//			const string&	prompt,		// prompt to user
//			T				answer		// default answer
//		);
//
//		T RequestT(			// without default
//			const string&	prompt		// prompt to user
//		);
//
//		bool ReadingT(
//			const string&	prompt,		// prompt to user
//			T&				answer		// user answer
//		);
//
//		T StringToT(
//			const string&	source,	// string to read
//			int				start,	// start position to read
//			int&			next,	// next  position to read
//			int&			error	// error code
//		);
//
// For backward compatibility, we also retain the variation:
//
//		T StringToT(
//			char*			start,	// start position to read
//			char*&			next,	// next  position to read
//			int&			error	// error code
//		);
//
// The two functions  RequestT  prompt the user for data of type  T
// One supplies a default answer and the other insists on an answer
// from the user
//
// The function  ReadingT  returns  true  if the user provides data
// in which case the data is passed back in the parameter  answer
// If the user does not provide data then  ReadingT  returns  false
//
// The function  ReadingT  can be used in loops of the form:
//		while (ReadingT(...)) { ... }
// The loop terminates when the user declines to provide data
//
// The function  StringToT  parses the input string and converts it
// to type  T
//
// The  next  parameter indicates where to continue further parsing
//
// The  error  parameter returns a positive error value if there is
// a parsing failure
//
// The function  StringToDouble  is the basis for all other numeric
// conversions.  StringToDouble parses its input with an arithmetic
// expression evaluator which we describe below:
//
//
// Parsing Details:
//
// The string reader that converts to double parses fairly general
// arithmetic expressions using the operators: + - * / % #.  The
// rationale for the extra operator # is to support two forms of
// division:
// 
//		/	usual floating point division
//		#	floating division followed by truncate to an integer
//
// For example:
//
//		8 / 5   is 1.6
//		8 # 5   is 1.0
//		9.7 # 1 is 9.0
//
// When parsing integer types (int, long, short), the calculation is
// done using the double precision routine and is then rounded NOT
// truncated.  This is the only convention that makes sense for
// arbitrary input expressions. This implies that for integer types:
//
//		8 / 5 becomes 2
//		8 # 5 becomes 1
//
// This means that / does not work as in C and C++.  The problem is
// really in C and C++ since / is overloaded in an inconsistent way.
// To achieve the classic integer division on input use # not /.
//
// The parsing routine can also recognize certain function calls:
//
// Trigonometric functions in radians:
//
//		sin(x),  cos(x),  tan(x)
//		asin(x), acos(x), atan(x), atan2(y, x)
//
// Trigonometric functions in degrees:
//
//		sindeg(x),  cosdeg(x),  tandeg(x)
//		asindeg(x), acosdeg(x), atandeg(x), atan2deg(y, x)
//
// Hyperbolic functions:
//
//		sinh(x), cosh(x), tanh(x)
//
// Exponentals and logarithms:
//
//		exp(x), ln(x), log(x), log10(x), sqrt(x), pow(x, y)
//
// The functions ln(x) and log(x) are identical.
//
// The parsing routine also recognizes one constant: pi = 3.14159...
//
//
// Note that for all types  T  except  double  there is a conversion
// function:
//
//		bool DoubleToT(double d, T& x);
//
// The converted value is stored in x.  If the conversion is correct
// then DoubleToT returns true.  If not, it returns false.


// Double precision floats

double RequestDouble(		// with default
	const string&	prompt,			// prompt to user
	double			answer			// default answer
);

double RequestDouble(		// without default
	const string&	prompt			// prompt to user
);

bool ReadingDouble(
	const string&	prompt,			// prompt to user
	double&			answer			// user answer if provided
);

double StringToDouble(
	const string&	source,			// string to read
	int				start,			// start position to read
	int&			next,			// next  position to read
	int&			error			// error code
);

double StringToDouble(
	const char*		start,			// start position to read
	const char*&	next,			// next  position to read
	int&			error			// error code
);


// Single precision floats

float RequestFloat(			// with default
	const string&	prompt,			// prompt to user
	float			answer			// default answer
);

float RequestFloat(			// without default
	const string&	prompt			// prompt to user
);

bool ReadingFloat(
	const string&	prompt,			// prompt to user
	float&			answer			// user answer if provided
);

float StringToFloat(
	const string&	source,			// string to read
	int				start,			// start position to read
	int&			next,			// next  position to read
	int&			error			// error code
);

float StringToFloat(
	const char*		start,			// start position to read
	const char*&	next,			// next  position to read
	int&			error			// error code
);

bool DoubleToFloat(double d, float& x);


// Integer: int

int RequestInt(				// with default
	const string&	prompt,			// prompt to user
	int				answer			// default answer
);

int RequestInt(				// without default
	const string&	prompt			// prompt to user
);

bool ReadingInt(
	const string&	prompt,			// prompt to user
	int&			answer			// user answer if provided
);

int StringToInt(
	const string&	source,			// string to read
	int				start,			// start position to read
	int&			next,			// next  position to read
	int&			error			// error code
);

int StringToInt(
	const char*		start,			// start position to read
	const char*&	next,			// next  position to read
	int&			error			// error code
);

bool DoubleToInt(double d, int& x);


// Integer: long

long RequestLong(			// with default
	const string&	prompt,			// prompt to user
	long			answer			// default answer
);

long RequestLong(			// without default
	const string&	prompt			// prompt to user
);

bool ReadingLong(
	const string&	prompt,			// prompt to user
	long&			answer			// user answer if provided
);

long StringToLong(
	const string&	source,			// string to read
	int				start,			// start position to read
	int&			next,			// next  position to read
	int&			error			// error code
);

long StringToLong(
	const char*		start,			// start position to read
	const char*&	next,			// next  position to read
	int&			error			// error code
);

bool DoubleToLong(double d, long& x);


// Integer: short

short RequestShort(			// with default
	const string&	prompt,			// prompt to user
	short			answer			// default answer
);

short RequestShort(			// without default
	const string&	prompt			// prompt to user
);

bool ReadingShort(
	const string&	prompt,			// prompt to user
	short&			answer			// user answer if provided
);

short StringToShort(
	const string&	source,			// string to read
	int				start,			// start position to read
	int&			next,			// next  position to read
	int&			error			// error code
);

short StringToShort(
	const char*		start,			// start position to read
	const char*&	next,			// next  position to read
	int&			error			// error code
);

bool DoubleToShort(double d, short& x);


// For unsigned types will only support unsigned long

unsigned long RequestUnsignedLong(	// with default
	const string&	prompt,			// prompt to user
	unsigned long	answer			// default answer
);

unsigned long RequestUnsignedLong(	// without default
	const string&	prompt			// prompt to user
);

bool ReadingUnsignedLong(
	const string&	prompt,			// prompt to user
	unsigned long&	answer			// user answer if provided
);

unsigned long StringToUnsignedLong(
	const string&	source,			// string to read
	int				start,			// start position to read
	int&			next,			// next  position to read
	int&			error			// error code
);

unsigned long StringToUnsignedLong(
	const char*		start,			// start position to read
	const char*&	next,			// next  position to read
	int&			error			// error code
);

bool DoubleToUnsignedLong(double d, unsigned long& x);


///////////////////////////////////////////////////////////////////////////////

// Number to string format routines

// Foundation routines will format to a string stream
// These routines will be applied to format to a string or to a CharBuffer (for
// backward compatibility)

// Case 1: Arguments: ... double x
//
// if x is negative then output a minus sign and make x positive
// if x equals an unsigned long integer then format x as integer
// otherwise format x as double using:
//		precision = 6
//		fixed format if 0.1 <= x <= 1000000 otherwise scientific

// Case 2: Arguments: ... double x, int width, int precision, bool fixed = true
//
// format double x using specified field width and decimal precision
// if fixed is true  use fixed format and right justify
// if fixed is false use scientific format and left justify
// if width > 0      add blank if needed to make sure that decimal points align

// Case 3: Arguments: ... long x, int width = 0
//
// if width > 0 then
//		format x using field width and right justify
// otherwise
//		format x using minimal space

// Case 4: Arguments: ... unsigned long x, int width = 0
//
// if width > 0 then
//		format x using field width and right justify
// otherwise
//		format x using minimal space

///////////////////////////////////////////////////////////////////////////////

// Format using stringstream

void Format(stringstream& stream, double x);

void Format(stringstream& stream, double x, int width, int precision,
		bool fixed = true);

void Format(stringstream& stream, long x, int width = 0);

void Format(stringstream& stream, unsigned long x, int width = 0);

///////////////////////////////////////////////////////////////////////////////

// Format using string

void Format(string& target, double x);

void Format(string& target, double x, int width, int precision,
		bool fixed = true);

void Format(string& target, long x, int width = 0);

void Format(string& target, unsigned long x, int width = 0);

///////////////////////////////////////////////////////////////////////////////

// Format using a SmallBuffer to provide storage for the string

char* Format(SmallBuffer& B, double x);

char* Format(SmallBuffer& B, double x, int width, int precision,
		bool fixed = true);

char* Format(SmallBuffer& B, long x, int width = 0);

char* Format(SmallBuffer& B, unsigned long x, int width = 0);

///////////////////////////////////////////////////////////////////////////////

// Routine to print double to cout using the rules of Format Case 1 above
// Used to print default in numeric input routines

void PrintDouble(double x);

///////////////////////////////////////////////////////////////////////////////

// Error codes for numeric data input

enum {
	NoError,				// No error = 0
	NumberError,			// Error: Reading numeric data
	ColorError,				// Error: Reading color data
	SymbolError,			// Error: Undefined identifier
	ConvertError,			// Error: Out of bounds
	AddError,				// Error: Addition
	SubError,				// Error: Subtraction
	MulError,				// Error: Multiplication
	DivError,				// Error: Division
	ModError,				// Error: Modulus
	NoLPError,				// Error: No starting (
	NoRPError,				// Error: No matching )
	NoCommaError,			// Error: Comma required
	NoBARError,				// Error: No matching |
	NoDataError,			// Error: Data required
	BadDataError,			// Error: Unexpected data
	EndDataError,			// Error: Data ends prematurely
	FunctionEvalError		// Error: Function evaluation error
};

// Obtain error message from error code

const char* IOErrorMessage(int error);

// Print error message and signal error position in string
// Request repeat of input data

void ErrorPrint(const string& source, int next, int error);

void ErrorPrint(const char* start, const char* next, int error);

///////////////////////////////////////////////////////////////////////////////

// Console control
// Omit: Summer 97
// void ConsoleTextJustify(int how);

#endif // IOTOOLS_H_
