// Copyright 1997
// College of Computer Science
// Northeastern University Boston MA 02115

// This software may be used for educational purposes as long as this copyright
// notice is retained at the top of all files

// Should this software be modified, the words "Modified from Original" must be
// included as a comment below this notice

// All publication rights are retained.  This software or its documentation may
// not be published in any media either in whole or in part.

///////////////////////////////////////////////////////////////////////////////

// MathUtil.h

///////////////////////////////////////////////////////////////////////////////

#ifndef MATHUTIL_H_
#define MATHUTIL_H_

#include <limits>						// For numeric limits
#include <math.h>						// For mathematical functions


// Useful numeric limits

const double			maximum_double			= numeric_limits<double>::			max();
const double			minimum_double			= - maximum_double;

const float				maximum_float			= numeric_limits<float>::			max();
const float				minimum_float			= - maximum_float;

const int				maximum_int				= numeric_limits<int>::				max();
const int				minimum_int				= numeric_limits<int>::				min();

const long				maximum_long			= numeric_limits<long>::			max();
const long				minimum_long			= numeric_limits<long>::			min();

const short				maximum_short			= numeric_limits<short>::			max();
const short				minimum_short			= numeric_limits<short>::			min();

const unsigned int		maximum_unsigned_int	= numeric_limits<unsigned int>::	max();

const unsigned long		maximum_unsigned_long	= numeric_limits<unsigned long>::	max();

const unsigned short	maximum_unsigned_short	= numeric_limits<unsigned short>::	max();


// Useful numeric test functions

inline bool ValidDouble(double x) { return (minimum_double <= x && x <= maximum_double); }

inline bool ValidFloat (float  x) { return (minimum_float  <= x && x <= maximum_float ); }


// Utility: SafeRoundToInt returns x safely rounded to an integer

inline int SafeRoundToInt(double d) {
	const double upper = maximum_int + 0.5;
	
	if (d >= upper)
		return maximum_int;
	
	const double lower = minimum_int - 0.5;
	
	if (d <= lower)
		return minimum_int;
	
	return round(d);
}


// Utility: SafeRoundToLong returns x safely rounded to a long integer

inline long SafeRoundToLong(double d) {
	const double upper = maximum_long + 0.5;
	
	if (d >= upper)
		return maximum_long;
	
	const double lower = minimum_long - 0.5;
	
	if (d <= lower)
		return minimum_long;
	
	return round(d);
}


// Utility: SafeRoundToShort returns x safely rounded to a short integer

inline short SafeRoundToShort(double d) {
	const double upper = maximum_short + 0.5;
	
	if (d >= upper)
		return maximum_short;
	
	const double lower = minimum_short - 0.5;
	
	if (d <= lower)
		return minimum_short;
	
	return round(d);
}


// Utility: SafeRoundToUnsignedLong returns x safely rounded to an unsigned long integer

inline unsigned long SafeRoundToUnsignedLong(double d) {
	const double upper = maximum_unsigned_long + 0.5;
	
	if (d >= upper)
		return maximum_unsigned_long;
	
	const double lower = - 0.5;
	
	if (d <= lower)
		return 0;
	
	return round(d);
}


// For backward compatibility, define SafeRound:

inline short SafeRound(double d) { return SafeRoundToShort(d); }


// Utility: Ratio returns the double precision ratio of 2 long integers

inline double Ratio (long x, long y) {
	double xx = x;
	double yy = y;
	
	return xx/yy;
}


// Utility: Expand interval from a to b to include x

inline void ExpandInterval(double& a, double& b, double x) {
	// put a and b into the right order
	if (b < a) {
		double c = b;
		b = a;
		a = c;
	}
	
	// expand limits
	if (x < a)
		a = x;
	else
	if (x > b)
		b = x;
}


// Radians to Degrees Conversions

const double pi = 3.14159265358979324;

const double radians_to_degrees = 180 / pi;

const double degrees_to_radians = pi / 180;


// Omit: Summer 97
// const double onedegree = pi/180;


// Trigonometric Functions in Degrees

inline double sindeg(double x) {
	return sin(x * degrees_to_radians);
}

inline double cosdeg(double x) {
	return cos(x * degrees_to_radians);
}

inline double tandeg(double x) {
	return tan(x * degrees_to_radians);
}

inline double asindeg(double x) {
	return asin(x) * radians_to_degrees;
}

inline double acosdeg(double x) {
	return acos(x) * radians_to_degrees;
}

inline double atandeg(double x) {
	return atan(x) * radians_to_degrees;
}

// notice that the two parameter form of the
// arc tangent takes y before x

inline double atan2deg(double y, double x) {
	return atan2(y, x) * radians_to_degrees;
}


// Natural Logarithm Synonym

inline double ln(double x) { return log(x); }

#endif // MATHUTIL_H_