// Copyright 1997
// College of Computer Science
// Northeastern University Boston MA 02115

// This software may be used for educational purposes as long as this copyright
// notice is retained at the top of all files

// Should this software be modified, the words "Modified from Original" must be
// included as a comment below this notice

// All publication rights are retained.  This software or its documentation may
// not be published in any media either in whole or in part.

///////////////////////////////////////////////////////////////////////////////

// Plot2D.cp

#include "Plot2D.h"

// Prototype for helper function for GuessDelta

void GuessOne(double a1, double a2, double& delta);


// Marks

void Mark(int mark, int size) {
	// get the current pen spot
	int x;
	int y;
	
	GetPenSpot(x, y);
	
	// sanity check for size
	if (size <= 0)
		size = 1;
	
	// draw mark
	switch (mark) {
		case DotMark:			// simple dot
			LineTo(x,y);
			break;
			
		case HBarMark:			// horizontal bar: can be used for vertical axis tick
			DrawLine(x - size, y, x + size, y);
			break;
			
		case VBarMark:			// vertical bar: can be used for horizontal axis tick
			DrawLine(x, y - size, x, y + size);
			break;
			
		case PlusMark:			// plus:  +
			DrawLine(x - size, y, x + size, y);
			DrawLine(x, y - size, x, y + size);
			break;
			
		case CrossMark:			// cross: x
			DrawLine(x - size, y - size, x + size, y + size);
			DrawLine(x - size, y + size, x + size, y - size);
			break;
			
		case StarMark:			// star:  *
			DrawLine(x - size, y, x + size, y);
			DrawLine(x, y - size, x, y + size);
			DrawLine(x - size, y - size, x + size, y + size);
			DrawLine(x - size, y + size, x + size, y - size);
			break;
			
		case SquareMark:		// open square
			// add 1 so right and bottom boundaries are correct
			FrameRect(x - size, y - size, x + size + 1, y + size + 1);
			break;
			
		case CircleMark:		// open circle
			// add 1 so right and bottom boundaries are correct
			FrameOval(x - size, y - size, x + size + 1, y + size + 1);
			break;
			
		case WedgeUMark:		// triangle wedge up
			MoveTo(x, y - size);
			LineTo(x - size, y + size);
			LineTo(x + size, y + size);
			LineTo(x, y - size);
			break;
			
		case WedgeDMark:		// triangle wedge down
			MoveTo(x, y + size);
			LineTo(x + size, y - size);
			LineTo(x - size, y - size);
			LineTo(x, y + size);
			break;
			
		case WedgeLMark:		// triangle wedge left
			MoveTo(x - size, y);
			LineTo(x + size, y + size);
			LineTo(x + size, y - size);
			LineTo(x - size, y);
			break;
			
		case WedgeRMark:		// triangle wedge right
			MoveTo(x + size, y);
			LineTo(x - size, y - size);
			LineTo(x - size, y + size);
			LineTo(x + size, y);
			break;
			
		case FillSquareMark:	// filled square
			// add 1 so right and bottom boundaries are correct
			PaintRect(x - size, y - size, x + size + 1, y + size + 1);
			break;
			
		case FillCircleMark:	// filled circle
			// add 1 so right and bottom boundaries are correct
			PaintOval(x - size, y - size, x + size + 1, y + size + 1);
			break;
		
		case FillWedgeUMark:	// filled triangle wedge up
			{
				int i = 0;
				int k = 0;
				int z = y - size;
				int limit = 2 * size;
				
				for (; i <= limit; i++, k = (i + 1)/2, z++)					
					DrawLine(x - k, z, x + k, z);
			}
			break;
		
		case FillWedgeDMark:	// filled triangle wedge down
			{
				int i = 0;
				int k = 0;
				int z = y + size;
				int limit = 2 * size;
				
				for (; i <= limit; i++, k = (i + 1)/2, z--)					
					DrawLine(x - k, z, x + k, z);
			}
			break;
		
		case FillWedgeLMark:	// filled triangle wedge left
			{
				int i = 0;
				int k = 0;
				int z = x - size;
				int limit = 2 * size;
				
				for (; i <= limit; i++, k = (i + 1)/2, z++)					
					DrawLine(z, y - k, z, y + k);
			}
			break;
		
		case FillWedgeRMark:	// filled triangle wedge right
			{
				int i = 0;
				int k = 0;
				int z = x + size;
				int limit = 2 * size;
				
				for (; i <= limit; i++, k = (i + 1)/2, z--)					
					DrawLine(z, y - k, z, y + k);
			}
			break;
		
		default:
			break;
	} // end switch to draw mark
	
	// restore the current pen spot
	MoveTo(x, y);
}


// Data plot and mark routines

void PlotData(
	const Point2D data[ ],			// data array
	long validsize,					// amount of valid data in array
	const LinearScale2D& T			// scaling transform from world to image
) {
	if (validsize > 0) {
		// move to initial data point
		MoveTo(T(data[0]));
		
		// draw lines to successive data points
		for (long i = 0; i < validsize; i++)
			LineTo(T(data[i]));
	}
}


void MarkData(
	const Point2D data[ ],			// data array
	long validsize,					// amount of valid data in array
	const LinearScale2D& T,			// scaling transform from world to image
	int mark,						// mark code
	int size						// size
) {
	for (long i = 0; i < validsize; i++) {
		MoveTo(T(data[i]));
		Mark(mark, size);
	}
}


// Axes routines

void PlotAxes(
	const Rect2D& bounds,			// data bounds
	const LinearScale2D& T			// scaling transform from world to image
) {
	// bounds coordinates
	double x1;
	double y1;
	double x2;
	double y2;
	
	bounds.Get(x1, y1, x2, y2);
	
	// origin coordinates
	double x;
	double y;
	
	if (x1 <= 0 && 0 <= x2)
		x = 0;						// if in bounds put axis at x = 0
	else
		x = x1;						// otherwise at x = x1
	
	if (y1 <= 0 && 0 <= y2)
		y = 0;						// if in bounds put axis at y = 0
	else
		y = y1;						// otherwise at y = y1
	
	// axis endpoints
	Point2D P;
	Point2D Q;
	
	// x axis
	P.Set(x1, y);
	Q.Set(x2, y);
	
	MoveTo(T(P));
	LineTo(T(Q));
	
	// y axis
	P.Set(x, y1);
	Q.Set(x, y2);
	
	MoveTo(T(P));
	LineTo(T(Q));
}


void MarkAxes(
	const Rect2D& bounds,			// data bounds
	const LinearScale2D& T,			// scaling transform from world to image
	double xdelta,					// spacing on x axis
	double ydelta,					// spacing on y axis
	int size						// hash mark size
) {
	// bounds coordinates
	double x1;
	double y1;
	double x2;
	double y2;
	
	bounds.Get(x1, y1, x2, y2);
	
	// origin coordinates
	double x;
	double y;
	
	if (x1 <= 0 && 0 <= x2)
		x = 0;						// if in bounds put axis at x = 0
	else
		x = x1;						// otherwise at x = x1
	
	if (y1 <= 0 && 0 <= y2)
		y = 0;						// if in bounds put axis at y = 0
	else
		y = y1;						// otherwise at y = y1
	
	// mark position and coordinates
	Point2D P;
	double xp;
	double yp;
	
	// loop variables
	short min;
	short max;
	short index;
	
	// along x axis
	if (xdelta > 0) {
		min = SafeRoundToShort(x1 / xdelta);
		
		if (min * xdelta < x1)
			min++;
		
		max = SafeRoundToShort(x2 / xdelta);
		
		if (max * xdelta > x2)
			max--;
		
		for (index = min; index <= max; index++) {
			xp = index * xdelta;
			P.Set(xp, y);
			
			MoveTo(T(P));
			Mark(VBarMark, size);
		}
	}
	
	// along y axis
	if (ydelta > 0) {
		min = SafeRoundToShort(y1 / ydelta);
		
		if (min * ydelta < y1)
			min++;
		
		max = SafeRoundToShort(y2 / ydelta);
		
		if (max * ydelta > y2)
			max--;
		
		for (index = min; index <= max; index++) {
			yp = index * ydelta;
			P.Set(x, yp);
			
			MoveTo(T(P));
			Mark(HBarMark, size);
		}
	}
}


void PlotGrids(
	const Rect2D& bounds,			// data bounds
	const LinearScale2D& T,			// scaling transform from world to image
	double xdelta,					// spacing on x axis
	double ydelta					// spacing on y axis
) {
	// bounds coordinates
	double x1;
	double y1;
	double x2;
	double y2;
	
	bounds.Get(x1, y1, x2, y2);
	
	// grid positions and coordinates
	Point2D P;
	Point2D Q;
	double xp;
	double yp;
	
	// loop variables
	short min;
	short max;
	short index;
	
	// vertical grid
	if (xdelta > 0) {
		min = SafeRoundToShort(x1 / xdelta);
		
		if (min * xdelta < x1)
			min++;
		
		max = SafeRoundToShort(x2 / xdelta);
		
		if (max * xdelta > x2)
			max--;
		
		for (index = min; index <= max; index++) {
			xp = index * xdelta;
			P.Set(xp, y1);
			Q.Set(xp, y2);
			
			MoveTo(T(P));
			LineTo(T(Q));
		}
	}
	
	// horizontal grid
	if (ydelta > 0) {
		min = SafeRoundToShort(y1 / ydelta);
		
		if (min * ydelta < y1)
			min++;
		
		max = SafeRoundToShort(y2 / ydelta);
		
		if (max * ydelta > y2)
			max--;
		
		for (index = min; index <= max; index++) {
			yp = index * ydelta;
			P.Set(x1, yp);
			Q.Set(x2, yp);
			
			MoveTo(T(P));
			LineTo(T(Q));
		}
	}
}


void GuessOne(double a1, double a2, double& delta) {
	if (a2 <= a1) {
		delta = 0;
		return;
	}
	
	// helper variables
	
	double difference = a2 - a1;
	double ratio;

	// find largest power of 10 < difference
	
	delta = 1;
	
	while (delta <  difference)		// go up
		delta *= 10;
	
	while (delta >= difference)		// go down
		delta /= 10;
	
	// now delta has the form 10^k and
	// delta < difference <= 10 * delta
	
	// get ratio of difference to delta
	// will have: 1 < ratio <= 10
	
	ratio = difference / delta;
	
	// reduce delta if ratio is less than 6
	
	if (ratio < 6) {
		if (ratio >= 3)
			delta /= 2;		// delta becomes 5 * 10^(k-1)
		else
			delta /= 5;		// delta becomes 2 * 10^(k-1)
	}
}


void GuessDelta(
	const Rect2D& bounds,			// data bounds
	double& xdelta,					// spacing on x axis
	double& ydelta					// spacing on y axis
) {
	// bounds coordinates
	double x1;
	double y1;
	double x2;
	double y2;
	
	bounds.Get(x1, y1, x2, y2);
	
	GuessOne(x1, x2, xdelta);
	GuessOne(y1, y2, ydelta);
}


// In FindBounds, assume that bounds rectangle is uninitialized

void FindBounds(
	Rect2D& bounds,					// bounds rectangle to calculate
	const Point2D data[ ],			// data array
	long validsize					// amount of valid data in array
) {
	// initialize bounds
	Point2D P;						// P initialized to origin
	
	if (validsize > 0)				// if data exists, reset
		P = data[0];				// P to first data point
	
	bounds.Set(P, P);				// set bounds to include P
	
	// now handle remaining points
	GrowBounds(bounds, data, validsize);
}

// In GrowBounds, assume that bounds rectangle is initialized with earlier
// data and that goal is to expand bounds to include additional data

void GrowBounds(
	Rect2D& bounds,					// bounds rectangle to calculate
	const Point2D data[ ],			// data array
	long validsize					// amount of valid data in array
) {
	for (long i = 0; i < validsize; i++)
		bounds.Expand(data[i]);
}

