// Copyright 1997
// College of Computer Science
// Northeastern University Boston MA 02115

// This software may be used for educational purposes as long as this copyright
// notice is retained at the top of all files

// Should this software be modified, the words "Modified from Original" must be
// included as a comment below this notice

// All publication rights are retained.  This software or its documentation may
// not be published in any media either in whole or in part.

///////////////////////////////////////////////////////////////////////////////

// Plot2D.h

// Tools are provided to create and plot data arrays with lines and/or markers
// using a specifed LinearScale2D

///////////////////////////////////////////////////////////////////////////////

#ifndef PLOT2D_H_
#define PLOT2D_H_

#include "Scale2D.h"

// Marks

enum Marks {
	// simple open marks
	DotMark,			// simple dot
	HBarMark,			// horizontal bar: can be used for vertical axis tick
	VBarMark,			// vertical bar: can be used for horizontal axis tick
	PlusMark,			// plus:  +
	CrossMark,			// cross: x
	StarMark,			// star:  *
	
	// simple closed marks
	SquareMark,			// open square
	FillSquareMark,		// filled square
	
	CircleMark,			// open circle
	FillCircleMark,		// filled circle
	
	WedgeUMark,			// triangle wedge up
	FillWedgeUMark,		// filled triangle wedge up
	
	WedgeDMark,			// triangle wedge down
	FillWedgeDMark,		// filled triangle wedge down
	
	WedgeLMark,			// triangle wedge left
	FillWedgeLMark,		// filled triangle wedge left
	
	WedgeRMark,			// triangle wedge right
	FillWedgeRMark		// filled triangle wedge right
};


enum {
	default_mark_size = 3
};


// Fundamental mark routine
//		Draws the mark in the desired size at the current graphics location
//		and maintains that location for subsequent graphics

void Mark(int mark, int size = default_mark_size);


// Data plot and mark routines

void PlotData(
	const Point2D data[ ],			// data array
	long validsize,					// amount of valid data in array
	const LinearScale2D& T			// scaling transform from world to image
);


void MarkData(
	const Point2D data[ ],			// data array
	long validsize,					// amount of valid data in array
	const LinearScale2D& T,			// scaling transform from world to image
	int mark,						// mark code
	int size = default_mark_size	// size
);


// Axes routines

void PlotAxes(
	const Rect2D& bounds,			// data bounds
	const LinearScale2D& T			// scaling transform from world to image
);


// Note: In MarkAxes and PlotGrids, if xdelta <= 0 and/or ydelta <= 0 then
// that direction is skipped 

void MarkAxes(
	const Rect2D& bounds,			// data bounds
	const LinearScale2D& T,			// scaling transform from world to image
	double xdelta,					// spacing on x axis
	double ydelta,					// spacing on y axis
	int size = default_mark_size	// size
);


void PlotGrids(
	const Rect2D& bounds,			// data bounds
	const LinearScale2D& T,			// scaling transform from world to image
	double xdelta,					// spacing on x axis
	double ydelta					// spacing on y axis
);


// Routine to guess good values for xdelta and ydelta
// for use in MarkAxes and PlotGrids
// Will seek to make each delta have the form f * 10^k
// where f = 1, 2, 5 and k is a short

void GuessDelta(
	const Rect2D& bounds,			// data bounds
	double& xdelta,					// spacing on x axis
	double& ydelta					// spacing on y axis
);


// Routines to build data tables from function definitions

// All routines to build data tables need to know:
//		data		array to store the Point2D data
//		validsize	number of data points to store
//		lower		lower bound of parameter range
//		upper		upper bound of parameter range
// The data array must have enough space to hold validsize points!

// Two cases:
//
//		Ordinary curve:			y = F(x)
//		Parametrized curve:		x = F(t)	y = G(t)
//
// These are distinguished by number of function paramters passed!


template <class DoubleFunction>
void BuildTable(
	Point2D	data[ ],				// array to store the Point2D data
	long validsize,					// number of data points to store
	double lower,					// lower bound of parameter range
	double upper,					// upper bound of parameter range
	const DoubleFunction& F			// function for y = F(x)
) {
	// return if there is no array data to calculate
	if (validsize <= 0)
		return;
	
	// fill in first position
	data[0].Set(lower, F(lower));
	
	// return if there is only one data point
	if (validsize == 1)
		return;
	
	// n is the number of subdivisions between lower and upper
	long n = validsize - 1;
	
	// delta is the size of these subdivisions
	double delta = (upper - lower) / double(n);
	
	// x is a position variable
	double x;
	
	// fill in intermediate positions
	for (long i = 1; i < n; i++) {
		x = lower + i * delta;
		data[i].Set(x, F(x));
	}
	
	// fill in final position
	data[n].Set(upper, F(upper));
}


template <class DoubleFunction>
void BuildTable(
	Point2D data[ ],				// array to store the Point2D data
	long validsize,					// number of data points to store
	double lower,					// lower bound of parameter range
	double upper,					// upper bound of parameter range
	const DoubleFunction& F,		// function for x = F(t)
	const DoubleFunction& G			// function for y = G(t)
) {
	// return if there is no array data to calculate
	if (validsize <= 0)
		return;
	
	// fill in first position
	data[0].Set(F(lower), G(lower));
	
	// return if there is only one data point
	if (validsize == 1)
		return;
	
	// n is the number of subdivisions between lower and upper
	long n = validsize - 1;
	
	// delta is the size of these subdivisions
	double delta = (upper - lower) / double(n);
	
	// t is a parameter variable
	double t;
	
	// fill in intermediate positions
	for (long i = 1; i < n; i++) {
		t = lower + i * delta;
		data[i].Set(F(t), G(t));
	}
	
	// fill in final position
	data[n].Set(F(upper), G(upper));
}


// Data bounds routines for use in defining scaling transform

// In FindBounds, assume that bounds rectangle is uninitialized

void FindBounds(
	Rect2D& bounds,					// bounds rectangle to calculate
	const Point2D data[ ],			// data array
	long validsize					// amount of valid data in array
);


// In GrowBounds, assume that bounds rectangle is initialized with earlier
// data and that goal is to expand bounds to include additional data

void GrowBounds(
	Rect2D& bounds,					// bounds rectangle to calculate
	const Point2D data[ ],			// data array
	long validsize					// amount of valid data in array
);


#endif // PLOT2D_H_
