// Copyright 1996
// College of Computer Science
// Northeastern University Boston MA 02115

// This software may be used for educational purposes as long as this copyright
// notice is retained at the top of all files

// Should this software be modified, the words "Modified from Original" must be
// included as a comment below this notice

// All publication rights are retained.  This software or its documentation may
// not be published in any media either in whole or in part.

///////////////////////////////////////////////////////////////////////////////

// SWindows.cp

#include "CHeaders.h"
#include "SWindows.h"


// Helper class to insure closure of drawing window

class WindowHelper {

public:
	WindowHelper() { };
	
	~WindowHelper() { CloseDrawing(); };
	
} theWindowHelper;


// Globals visible externally

int WindowMinX;							// minimum x for window corner
int WindowMinY;							// minimum y for window corner
int WindowMaxX;							// maximum x for window corner
int WindowMaxY;							// maximum y for window corner

int ConsoleCols;						// number of cols in console
int ConsoleRows;						// number of rows in console

int DrawMinX;							// minimum x in drawing window
int DrawMinY;							// minimum y in drawing window
int DrawMaxX;							// maximum x in drawing window
int DrawMaxY;							// maximum y in drawing window

Rect theDrawingRect;					// drawing window region

CWindowRecord	theDrawing;				// drawing window data structure
WindowPtr		theDrawingPtr = 0;		// pointer to drawing window data

// Globals visible internally

int SystemInitDone = 0;					// checks if SystemInit has been called

const int NormalMaxX = 640;				// x limit for standardized screen area
const int NormalMaxY = 480;				// y limit for standardized screen area

const int MinCols = 10;					// minimum cols in console
const int MinRows =  2;					// minimum rows in console

const int BigSize = 400;				// size limit for big square pair window

const int TinySize = 200;				// height limit for certain tiny windows
const int TinyColsLimit = 40;			// cols limit for certain tiny windows
const int TinyRowsLimit =  5;			// rows limit for certain tiny windows

const int StatusMaxSize = LargeFont;	// determines when to show status line
const int FontSizeLimit = 18;			// fontSize limit for smaller consoles

const int padding	=  8;				// padding allowance in sizing consoles
const int scrollbar	= 15;				// allowance for scrollbar width
const int titlebar	= 20;				// allowance for titlebar height

	// The following constants represent font dimensions for monaco fonts
	// The font sizes 9, 10, 12, 14, 18, 24, 48 are the ones supported in 
	// MakeConsole, BigConsole, MaxConsole, TinyConsole, and all pair functions

const int wid9  =  6;
const int wid10 =  6;
const int wid12 =  7;
const int wid14 =  8;
const int wid18 = 11;
const int wid24 = 14;
const int wid48 = 29;

const int hgt9  = 11;
const int hgt10 = 13;
const int hgt12 = 16;
const int hgt14 = 18;
const int hgt18 = 24;
const int hgt24 = 32;
const int hgt48 = 64;


// internal function prototypes

	// adjust window bounds to within full screen
	// handle default values passed as zero
	// check for SystemInit
void AdjustBounds(int& minX, int& minY, int& maxX, int& maxY);

	// adjust fontSize to 9,10,12,14,18,24,48
void AdjustFontSize(int& fontSize);

	// force fontSize <= 18
void LimitFontSize(int& fontSize);

	// find font dimensions
void FontDimensions(int fontSize, int& wid, int& hgt);


// function definitions


void AdjustBounds(int& minX, int& minY, int& maxX, int& maxY) {
	// initialize system if it has not yet been done
	SystemInit();
	
	// check for valid screen dimensions
	// adjust if passed parameter is too small or too large

	if ((minX < WindowMinX) || (minX >= WindowMaxX))
		minX = WindowMinX;

	if ((minY < WindowMinY) || (minY >= WindowMaxY))
		minY = WindowMinY;

	if ((maxX <= minX) || (maxX > WindowMaxX))
		maxX = WindowMaxX;

	if ((maxY <= minY) || (maxY > WindowMaxY))
		maxY = WindowMaxY;
}


void AdjustFontSize(int& fontSize) {
	if (fontSize > 24)
		fontSize = 48;
	else if (fontSize > 18)
		fontSize = 24;
	else if (fontSize > 14)
		fontSize = 18;
	else if (fontSize > 12)
		fontSize = 12;
	else if (fontSize > 10)
		fontSize = 12;
	else if (fontSize < 9)
		fontSize = 9;
}


void LimitFontSize(int& fontSize) {
	if (fontSize > FontSizeLimit)
		fontSize = FontSizeLimit;
}


void FontDimensions(int fontSize, int& wid, int& hgt) {
	switch (fontSize) {
		case 9:
			wid = wid9;
			hgt = hgt9;
			break;
		
		case 10:
			wid = wid10;
			hgt = hgt10;
			break;
		
		case 12:
			wid = wid12;
			hgt = hgt12;
			break;
		
		case 14:
			wid = wid14;
			hgt = hgt14;
			break;
		
		case 18:
			wid = wid18;
			hgt = hgt18;
			break;
		
		case 24:
			wid = wid24;
			hgt = hgt24;
			break;
			
		default:
			wid = wid48;
			hgt = hgt48;
	}
}


void SystemInit(bool handleEvents) {
	if (!SystemInitDone) {
		// Normal initialization
		
		MaxApplZone();
		MoreMasters();
		
		InitGraf(&qd.thePort);
		InitFonts();
		InitWindows();
		InitMenus();
		TEInit();
		InitDialogs(0L);
		
		FlushEvents(everyEvent, 0);
		
		InitCursor();
		
		// Set screen dimensions
		WindowMinX =  0;
		WindowMinY = 38;
		WindowMaxX = qd.screenBits.bounds.right;
		WindowMaxY = qd.screenBits.bounds.bottom;

		// Set certain console options
		
		SIOUXSettings.initializeTB		= FALSE;
		SIOUXSettings.standalone		= handleEvents;
		SIOUXSettings.setupmenus		= TRUE;
		
		SIOUXSettings.autocloseonquit	= FALSE;
		SIOUXSettings.asktosaveonclose	= FALSE;
		
		// Flag that initialization has been done
		
		SystemInitDone = 1;
	}
}


void OpenConsole(int fontSize, int cols, int rows, int minX, int minY) {
	// adjust window bounds
	// maxX and maxY are needed for call but are unused in this function
	
	int maxX = minX;
	int maxY = minY;
	
	AdjustBounds(minX, minY, maxX, maxY);
	
	// make sure console is not too small
	
	if (cols < MinCols)
		cols = MinCols;
	
	if (rows < MinRows)
		rows = MinRows;
	
	ConsoleCols = cols;
	ConsoleRows = rows;
	
	// set remaining console options
	
	SIOUXSettings.leftpixel	= minX;
	SIOUXSettings.toppixel	= minY;
	
	SIOUXSettings.columns	= cols;
	SIOUXSettings.rows		= rows;
	
	SIOUXSettings.fontid	= monaco;
	SIOUXSettings.fontsize	= fontSize;
	SIOUXSettings.fontface	= 0;
	SIOUXSettings.tabspaces	= 4;
	
	// show status line if fontSize <= StatusMaxSize
	
	if (fontSize <= StatusMaxSize)
		SIOUXSettings.showstatusline = TRUE;
	else
		SIOUXSettings.showstatusline = FALSE;
	
	// force console open with a newline!
	cout << endl;
}


void MakeConsole(
	int fontSize,
	int minX,
	int minY,
	int maxX,
	int maxY,
	int colsLimit,
	int rowsLimit)
{
	// adjust window bounds
	AdjustBounds(minX, minY, maxX, maxY);
	
	// adjust font size and find font width and height
	int wid;
	int hgt;
	
	AdjustFontSize(fontSize);
	FontDimensions(fontSize, wid, hgt);
	
	// compute maximum cols and rows in available space
	int availableX = maxX - minX - padding - scrollbar;
	int availableY = maxY - minY - padding;
	int cols = availableX / wid;
	int rows = availableY / hgt;
	
	// adjust for status line if fontSize <= StatusMaxSize
	
	if (fontSize <= StatusMaxSize) rows--;
	
	// adjust for cols and rows limits
	
	if (colsLimit > 0)
		if (cols > colsLimit)
			cols = colsLimit;
	
	if (rowsLimit > 0)
		if (rows > rowsLimit)
			rows = rowsLimit;
	
	// call OpenConsole to finish up
	
	OpenConsole(fontSize, cols, rows, minX, minY);
}


void BigConsole(int fontSize) {
	MakeConsole(fontSize, 0, 0, NormalMaxX, NormalMaxY);
}


void MaxConsole(int fontSize) {
	MakeConsole(fontSize);
}


void TinyConsole(int fontSize) {
	LimitFontSize(fontSize);
	
	// initialize system if it has not yet been done
	SystemInit();
	
	int minX = WindowMinX + TinySize + 1;
	int minY = WindowMinY;
	int maxX = WindowMaxX;
	int maxY = WindowMinY + TinySize;
	
	MakeConsole(fontSize, minX, minY, maxX, maxY, TinyColsLimit, 0);
}


void MakeDrawing(int minX, int minY, int maxX, int maxY) {
	// close earlier drawing window
	CloseDrawing();
	
	// adjust window bounds
	AdjustBounds(minX, minY, maxX, maxY);
	
	Rect bounds;							// set window bounds rectangle

	bounds.left			= minX;
	bounds.top			= minY;
	bounds.right		= maxX;
	bounds.bottom		= maxY;

	Str15 title			= "\pDrawing";		// window title
	short procID 		= noGrowDocProc;	// window style
	WindowPtr place		= WindowPtr(-1);	// position in front
	
	bool visible		= TRUE;				// make visible
	bool goAway			= TRUE;				// go away box
	
	long refCon			= 0;				// trivial refCon
	
	theDrawingPtr =
		NewCWindow(&theDrawing, &bounds, title, visible, procID, place, goAway, refCon);
	
	if (theDrawingPtr) {
		SelectDrawing();
		SetDrawingClip();
	}
	else
		cout << "Unable to open drawing window" << endl;
}


void BigDrawing(void) {
	MakeDrawing(0, 0, NormalMaxX, NormalMaxY);
}


void MaxDrawing(void) {
	MakeDrawing();
}


void TinyDrawing(void) {
	// initialize system if it has not yet been done
	SystemInit();
	
	int minX = WindowMinX;
	int minY = WindowMinY;
	
	int maxX = WindowMinX + TinySize;
	int maxY = WindowMinY + TinySize;
	
	MakeDrawing(minX, minY, maxX, maxY);
}


void MakeSquarePair(int fontSize, int minX, int minY, int maxX, int maxY) {
	int wid;
	int hgt;
	
	LimitFontSize(fontSize);
	AdjustFontSize(fontSize);
	FontDimensions(fontSize, wid, hgt);

	// adjust window bounds
	AdjustBounds(minX, minY, maxX, maxY);
	
	int deltaX = maxX - minX;
	int deltaY = maxY - minY;
	
	// make a square pair only if there is some space for a console on right
	// otherwise abort
	
	int spaceX = wid * MinCols + padding + scrollbar;
	int availableX = deltaX - spaceX;
	
	if (availableX > 0)
		if (availableX > deltaY)
			availableX = deltaY;
	
	if (availableX > 0) {
		// drawing window bounds
		int x1 = minX;
		int y1 = minY;
		int x2 = x1 + availableX;
		int y2 = y1 + availableX;
		
		// console window bounds
		int x3 = x2 + 1;
		int y3 = y1;
		int x4 = maxX;
		int y4 = maxY;
		
		MakeConsole(fontSize, x3, y3, x4, y4);
		MakeDrawing(x1, y1, x2, y2);
	}
	else
		cout << "Unable to make window pair with desired dimensions" << endl;
}


void BigSquarePair(int fontSize) {
	// initialize system if it has not yet been done
	SystemInit();
	
	// Make console portion of square pair
	LimitFontSize(fontSize);
	
	int minX = WindowMinX + BigSize + 1;
	int minY = WindowMinY;
	int maxX = NormalMaxX;
	int maxY = WindowMinY + BigSize;
	
	MakeConsole(fontSize, minX, minY, maxX, maxY, 0, 0);

	// Make drawing portion of square pair
	minX = 0;
	maxX = BigSize;
	MakeDrawing(minX, minY, maxX, maxY);
}


void MaxSquarePair(int fontSize) {
	MakeSquarePair(fontSize);
}


void TinySquarePair(int fontSize) {
	TinyConsole(fontSize);
	TinyDrawing();
}


void MakeWidePair(int fontSize, int minX, int minY, int maxX, int maxY) {
	int wid;
	int hgt;
	
	LimitFontSize(fontSize);
	AdjustFontSize(fontSize);
	FontDimensions(fontSize, wid, hgt);

	// adjust window bounds
	AdjustBounds(minX, minY, maxX, maxY);
	
	int deltaX = maxX - minX;
	int deltaY = maxY - minY;
	
	// make a wide pair only if there is some space for a console on bottom
	// otherwise abort
	
	// simple check for y space allowing for statusrow
	
	int spaceY;
	
	if (fontSize <= StatusMaxSize)
		spaceY = hgt * (MinRows + 1) + padding + titlebar;
	else
		spaceY = hgt * MinRows + padding + titlebar;
			
	int availableY = deltaY - spaceY;
	
	// simple check for y space in relation to x space
	// must be able to obtain 2-to-1 ratio
	
	if (availableY > 0) {
		if (deltaX < (2 * availableY))
			availableY = deltaX / 2;
	}
	
	if (availableY > 0) {
		
		// define size of largest reqion with 2-to-1 ratio
		int availableX = 2 * availableY;
				
		// drawing window bounds
		int x1 = minX;
		int y1 = minY;
		int x2 = x1 + availableX;
		int y2 = y1 + availableY;
		
		// console window bounds
		int x3 = x1;
		int y3 = y2 + titlebar + 1;
		int x4 = maxX;
		int y4 = maxY;
		
		MakeConsole(fontSize, x3, y3, x4, y4);
		MakeDrawing(x1, y1, x2, y2);
	}
	else
		cout << "Unable to make window pair with desired dimensions" << endl;
}


void BigWidePair(int fontSize) {
	MakeWidePair(fontSize, 0, 0, NormalMaxX, NormalMaxY);
}


void MaxWidePair(int fontSize) {
	MakeWidePair(fontSize);
}


void TinyWidePair(int fontSize) {
	LimitFontSize(fontSize);

	// initialize system if it has not yet been done
	SystemInit();
	
	// drawing window bounds
	int x1 = WindowMinX;
	int y1 = WindowMinY;
	int x2 = x1 + 2 * TinySize;
	int y2 = y1 + TinySize;
	
	// console window bounds
	int x3 = x1;
	int y3 = y2 + titlebar + 1;
	int x4 = WindowMaxX;
	int y4 = WindowMaxY;
	
	MakeConsole(fontSize, x3, y3, x4, y4, StdCols, TinyRowsLimit);
	MakeDrawing(x1, y1, x2, y2);
}


void CloseDrawing(void) {
	if (theDrawingPtr) {
		CloseWindow(theDrawingPtr);
		theDrawingPtr = 0;
	}
}


void ClearDrawing(void) {
	if (theDrawingPtr) {
		RgnHandle oldClipRgn;			// save clip region
		oldClipRgn = NewRgn();
		GetClip(oldClipRgn);
		
		SetDrawingClip();				// clear drawing frame
		EraseRect(&theDrawingRect);
		
		SetClip(oldClipRgn);			// restore clip region
		DisposeRgn(oldClipRgn);
	}
}


void SelectDrawing(void) {
	if (theDrawingPtr) {
		SetPort(theDrawingPtr);
		SelectWindow(theDrawingPtr);
	}
}


void SetDrawingRect(void) {
	if (theDrawingPtr) {		
		theDrawingRect = theDrawingPtr->portRect;
		
		DrawMinX = theDrawingRect.left;
		DrawMinY = theDrawingRect.top;
		DrawMaxX = theDrawingRect.right;
		DrawMaxY = theDrawingRect.bottom;
	}
}


void SetDrawingClip(void) {
	if (theDrawingPtr) {
		SetDrawingRect();		
		ClipRect(&theDrawingRect);
	}
}


// The following functions are not yet implemented

void SaveDrawWindow(void) {

}


void RefreshDrawing(void) {

}


