// Copyright 1996
// College of Computer Science
// Northeastern University Boston MA 02115

// This software may be used for educational purposes as long as this copyright
// notice is retained at the top of all files

// Should this software be modified, the words "Modified from Original" must be
// included as a comment below this notice

// All publication rights are retained.  This software or its documentation may
// not be published in any media either in whole or in part.

///////////////////////////////////////////////////////////////////////////////

// SWindows.h

#ifndef SIMPLEWINDOWS_H_
#define SIMPLEWINDOWS_H_

// comment line below since bool support enabled
// #include <bool.h>					// definition of bool, false, true

// Globals

const int SmallFont  =  9;				// common fontSize choices
const int MediumFont = 10;
const int LargeFont  = 12;

const int StdCols = 80;					// default columns in console
const int StdRows = 25;					// default rows    in console

const bool AutoHandleEvents = true;		// let console code handle events
const bool SelfHandleEvents = false;	// program will handle its events


// These globals computed by SystemInit

extern int WindowMinX;					// minimum x for window corner
extern int WindowMinY;					// minimum y for window corner
extern int WindowMaxX;					// maximum x for window corner
extern int WindowMaxY;					// maximum y for window corner

// These globals valid only when console window is open

extern int ConsoleCols;					// number of cols in console
extern int ConsoleRows;					// number of rows in console

// These globals valid only when drawing window is open

extern int DrawMinX;					// minimum x in drawing window
extern int DrawMinY;					// minimum y in drawing window
extern int DrawMaxX;					// maximum x in drawing window
extern int DrawMaxY;					// maximum y in drawing window

extern Rect theDrawingRect;				// drawing window region

extern CWindowRecord	theDrawing;		// drawing window data structure
extern WindowPtr		theDrawingPtr;	// pointer to drawing window data

///////////////////////////////////////////////////////////////////////////

// Notes on the function names for window creation:


// OpenConsole opens a text-io console window given the upper-left corner
// position of the window and the number of character columns and rows


// The other window creation functions combine a prefix and suffix to name
// one of 16 functions: 
//		Prefix: "Make", "Big", "Max", "Tiny"
//		Suffix: "Console", "Drawing", "SquarePair", "WidePair"


// "Make" style functions open the console and/or drawing window given the
// upper-left and lower-right position of the bounding rectangle that will
// enclose the window or the pair of windows
 
// "Big" style functions use a standard 640 by 480 screen area

// "Max" style functions use the entire available screen area

// "Tiny" style functions use tiny windows to save space on the screen for
// debugger windows


// "Console" style functions open a console window only

// "Drawing" style functions open a drawing window only

// "Pair" style functions open both a console and a drawing window

//   "SquarePair" makes the drawing area square

//   "WidePair" makes the drawing area have a 2-to-1 width-to-height ratio

///////////////////////////////////////////////////////////////////////////


// SystemInit does standard system initialization
// If handleEvents is TRUE then let console code handle events

// SystemInit is called by console and drawing window functions
// if it is not called explcitly by the user

void SystemInit(bool handleEvents = AutoHandleEvents);


// Text-IO Console WIndow Functions


// OpenConsole function
// Open console text-io window
// Parameters:
//		fontSize = point size of text font		default = 9
//		cols = number of text columns			default = 80
//		rows = number of text rows				default = 25
//		minX = left of text area				default to WindowMinX
//		minY = top  of text area				default to WindowMinY


void OpenConsole(
	int fontSize	= SmallFont,
	int cols		= StdCols,
	int rows		= StdRows,
	int minX		= 0,						// default to WindowMinX
	int minY		= 0							// default to WindowMinY
);


// MakeConsole function
// Make console by using window size limits and cols-rows limits
// Zero window size limits default to appropriate screen limits
// Zero limits on number of cols and rows default to no limits

void MakeConsole(
	int fontSize	= SmallFont,
	int minX		= 0,						// default to WindowMinX
	int minY		= 0,						// default to WindowMinY
	int maxX		= 0,						// default to WindowMaxX
	int maxY		= 0,						// default to WindowMaxY
	int colsLimit	= 0,						// default to no limit
	int rowsLimit	= 0							// default to no limit
);


// BigConsole function
// Open console window for 640 by 480 screen area

void BigConsole(int fontSize = SmallFont);


// MaxConsole function
// Open console window for full screen area

void MaxConsole(int fontSize = SmallFont);


// TinyConsole function
// Open console window with the following parameters:
//		Offset 200 pixels from left screen edge
//		Fit up to 40 text columns depending on space
//		Fit as many rows as will fit in 200 pixels height
//		fontSize will be restricted to maximum of 18 

void TinyConsole(int fontSize = SmallFont);


// Drawing Window Functions


// MakeDrawing function
// Open drawing window given window size limits
// Zero window size limits default to appropriate screen limits
// Parameters:
//		minX = left   of drawing area			// default to WindowMinX
//		minY = top    of drawing area			// default to WindowMinY
//		maxX = right  of drawing area			// default to WindowMaxX
//		maxY = bottom of drawing area			// default to WindowMaxY

void MakeDrawing(int minX = 0, int minY = 0, int maxX = 0, int maxY = 0);


// BigDrawing function
// Open drawing window for 640 by 480 screen area

void BigDrawing(void);


// MaxDrawing function
// Open drawing window for full screen area

void MaxDrawing(void);


// TinyDrawing function
// Open drawing window with 200 by 200 drawing region
// Position: Upper-left corner of screen

void TinyDrawing(void);


// Functions Which Open a Console-Drawing Window Pair
// Note: fontSize will be restricted to maximum of 18


// SquarePair functions
// Drawing region is square
// If possible:
// Place drawing window in square panel to the left of console window

// General function to position square pair in indicated region
void MakeSquarePair(
	int fontSize	= SmallFont,
	int minX		= 0,						// default to WindowMinX
	int minY		= 0,						// default to WindowMinY
	int maxX		= 0,						// default to WindowMaxX
	int maxY		= 0							// default to WindowMaxY
);

void BigSquarePair(int fontSize = SmallFont);
// Drawing region is 400 x 400 square

void MaxSquarePair(int fontSize = SmallFont);
// Drawing region is maximum square that fits in full screen

void TinySquarePair(int fontSize = SmallFont);
// Drawing region is 200 x 200 square


// WidePair functions
// Drawing region has width-to-height ratio of 2-to-1
// If possible:
// Place drawing window in wide panel above corresponding console panel

// General function to position wide pair in indicated region
void MakeWidePair(
	int fontSize	= SmallFont,
	int minX		= 0,						// default to WindowMinX
	int minY		= 0,						// default to WindowMinY
	int maxX		= 0,						// default to WindowMaxX
	int maxY		= 0							// default to WindowMaxY
);

void BigWidePair(int fontSize = SmallFont);
// Drawing region is maximum 2-to-1 panel that fits in 640 by 480 screen

void MaxWidePair(int fontSize = SmallFont);
// Drawing region is maximum 2-to-1 panel that fits in full screen

void TinyWidePair(int fontSize = SmallFont);
// Drawing region is 400 by 200


// Auxiliary Drawing Window Functions


// CloseDrawing function
// Close drawing window if it happens to be open

void CloseDrawing(void);


// ClearDrawing function
// Clear drawing window

void ClearDrawing(void);


// SelectDrawing function
// Make drawing window active

void SelectDrawing(void);


// find the Drawing area and set DrawMinX, DrawMinY, DrawMaxX, DrawMaxY
void SetDrawingRect(void);


// make the Drawing area the clipping region
void SetDrawingClip(void);


// SaveDrawWindow function
// Save contents of drawing window to off-screen copy

void SaveDrawWindow(void);


// RefreshDrawing function
// Refresh contents of drawing window from off-screen copy

void RefreshDrawing(void);


#endif // SIMPLEWINDOWS_H_